import numpy as np
import pytest
from astropy import units as u
from astropy.time import Time
from roman_datamodels import maker_utils
from roman_datamodels.datamodels import (
    GainRefModel,
    ImageModel,
    RampModel,
    ReadnoiseRefModel,
)
from roman_datamodels.dqflags import group

from romancal.ramp_fitting import RampFitStep

MAXIMUM_CORES = ["none", "quarter", "half", "all"]


rng = np.random.default_rng(619)


def test_ols_multicore_ramp_fit_match(make_data):
    """Test various core amount calculation"""
    model, override_gain, override_readnoise = make_data

    # read noise is also modified in place in an important way (!)
    # so we make copies here so that we can get agreement.
    out_model = RampFitStep.call(
        model.copy(),  # model1 is modified in place now.
        algorithm="ols",
        maximum_cores="none",
        override_gain=override_gain,
        override_readnoise=override_readnoise.copy(),
    )

    all_out_model = RampFitStep.call(
        model.copy(),  # model1 is modified in place now.
        algorithm="ols",
        maximum_cores="all",
        override_gain=override_gain,
        override_readnoise=override_readnoise.copy(),
    )

    # Original ramp parameters
    np.testing.assert_allclose(out_model.data, all_out_model.data, 1e-6, 1e-6)
    np.testing.assert_allclose(out_model.err, all_out_model.err, 1e-6, 1e-6)
    np.testing.assert_allclose(out_model.amp33, all_out_model.amp33, 1e-6, 1e-6)
    np.testing.assert_allclose(
        out_model.border_ref_pix_left, all_out_model.border_ref_pix_left, 1e-6, 1e-6
    )
    np.testing.assert_allclose(
        out_model.border_ref_pix_right, all_out_model.border_ref_pix_right, 1e-6, 1e-6
    )
    np.testing.assert_allclose(
        out_model.border_ref_pix_top, all_out_model.border_ref_pix_top, 1e-6, 1e-6
    )
    np.testing.assert_allclose(
        out_model.border_ref_pix_bottom, all_out_model.border_ref_pix_bottom, 1e-6, 1e-6
    )
    np.testing.assert_allclose(
        out_model.dq_border_ref_pix_left,
        all_out_model.dq_border_ref_pix_left,
        1e-6,
        1e-6,
    )
    np.testing.assert_allclose(
        out_model.dq_border_ref_pix_right,
        all_out_model.dq_border_ref_pix_right,
        1e-6,
        1e-6,
    )
    np.testing.assert_allclose(
        out_model.dq_border_ref_pix_top, all_out_model.dq_border_ref_pix_top, 1e-6, 1e-6
    )
    np.testing.assert_allclose(
        out_model.dq_border_ref_pix_bottom,
        all_out_model.dq_border_ref_pix_bottom,
        1e-6,
        1e-6,
    )

    # New rampfit parameters
    np.testing.assert_allclose(
        out_model.var_poisson, all_out_model.var_poisson, 1e-6, 1e-6
    )
    np.testing.assert_allclose(
        out_model.var_rnoise, all_out_model.var_rnoise, 1e-6, 1e-6
    )


@pytest.mark.parametrize("make_data", [(1, 1, 1, 20, 20)], indirect=True)
@pytest.mark.parametrize("max_cores", MAXIMUM_CORES)
def test_ols_one_group_small_buffer_fit(max_cores, make_data):
    model, override_gain, override_readnoise = make_data

    model.data[0, 15, 10] = 10.0 * model.data.unit  # add single CR
    override_gain.data[15, 10] = 2 * override_gain.data.unit

    out_model = RampFitStep.call(
        model,
        algorithm="ols",
        maximum_cores=max_cores,
        override_gain=override_gain,
        override_readnoise=override_readnoise,
    )

    data = out_model.data.value

    # Index changes due to trimming of reference pixels
    np.testing.assert_allclose(data[11, 6], 10, 1e-5)


@pytest.mark.parametrize("max_cores", MAXIMUM_CORES)
def test_ols_saturated_ramp_fit(max_cores, make_data):
    model, override_gain, override_readnoise = make_data

    # Set saturated flag
    model.groupdq = model.groupdq | group.SATURATED

    # Run ramp fit step
    out_model = RampFitStep.call(
        model,
        algorithm="ols",
        maximum_cores=max_cores,
        override_gain=override_gain,
        override_readnoise=override_readnoise,
    )

    # Test data and error arrays are zeroed out
    np.testing.assert_array_equal(out_model.data.value, 0)
    np.testing.assert_array_equal(out_model.err.value, 0)
    np.testing.assert_array_equal(out_model.var_poisson.value, 0)
    np.testing.assert_array_equal(out_model.var_rnoise.value, 0)

    # Test that all pixels are flagged saturated
    assert np.all(np.bitwise_and(out_model.dq, group.SATURATED) == group.SATURATED)

    # Test that original ramp parameters preserved
    np.testing.assert_allclose(out_model.amp33, model.amp33, 1e-6)
    np.testing.assert_allclose(
        out_model.border_ref_pix_left, model.border_ref_pix_left, 1e-6
    )
    np.testing.assert_allclose(
        out_model.border_ref_pix_right, model.border_ref_pix_right, 1e-6
    )
    np.testing.assert_allclose(
        out_model.border_ref_pix_top, model.border_ref_pix_top, 1e-6
    )
    np.testing.assert_allclose(
        out_model.border_ref_pix_bottom, model.border_ref_pix_bottom, 1e-6
    )
    np.testing.assert_allclose(
        out_model.dq_border_ref_pix_left, model.dq_border_ref_pix_left, 1e-6
    )
    np.testing.assert_allclose(
        out_model.dq_border_ref_pix_right, model.dq_border_ref_pix_right, 1e-6
    )
    np.testing.assert_allclose(
        out_model.dq_border_ref_pix_top, model.dq_border_ref_pix_top, 1e-6
    )
    np.testing.assert_allclose(
        out_model.dq_border_ref_pix_bottom, model.dq_border_ref_pix_bottom, 1e-6
    )

    # Test that an Image model was returned.
    assert type(out_model) == ImageModel

    # Test that the ramp fit step was labeled complete
    assert out_model.meta.cal_step.ramp_fit == "COMPLETE"


# ########
# fixtures
# ########
@pytest.fixture
def make_data(request):
    """Create test input data

    Parameters
    ----------
    request.param : (ingain, deltatime, ngroups, xsize, ysize)
        If specified, set the parameters of the created data.
        If not specified, defaults are used.

    Returns
    -------
    image, gain, readnoise : ImageModel, GainRefModel, ReadnoiseRefModel
        Input image and related references
    """
    if getattr(request, "param", None):
        ingain, deltatime, ngroups, xsize, ysize = request.param
    else:
        ingain = 1
        deltatime = 1
        ngroups = 4
        xsize = 20
        ysize = 20
    shape = (ngroups, xsize, ysize)

    image = generate_ramp_model(shape, deltatime)
    gain, readnoise = generate_wfi_reffiles(shape[1:], ingain)

    return image, gain, readnoise


# #########
# Utilities
# #########


def generate_ramp_model(shape, deltatime=1):
    data = u.Quantity(
        (rng.random(shape) * 0.5).astype(np.float32), u.DN, dtype=np.float32
    )
    err = u.Quantity(
        (rng.random(shape) * 0.0001).astype(np.float32), u.DN, dtype=np.float32
    )
    pixdq = np.zeros(shape=shape[1:], dtype=np.uint32)
    gdq = np.zeros(shape=shape, dtype=np.uint8)

    dm_ramp = maker_utils.mk_ramp(shape=shape)
    dm_ramp.data = u.Quantity(data, u.DN, dtype=np.float32)
    dm_ramp.pixeldq = pixdq
    dm_ramp.groupdq = gdq
    dm_ramp.err = u.Quantity(err, u.DN, dtype=np.float32)

    dm_ramp.meta.exposure.group_time = deltatime
    dm_ramp.meta.exposure.frame_time = deltatime
    dm_ramp.meta.exposure.ngroups = shape[0]
    dm_ramp.meta.exposure.nframes = 1
    dm_ramp.meta.exposure.groupgap = 0

    ramp_model = RampModel(dm_ramp)

    return ramp_model


def generate_wfi_reffiles(shape, ingain=6):
    # Create temporary gain reference file
    gain_ref = maker_utils.mk_gain(shape=shape)

    gain_ref["meta"]["instrument"]["detector"] = "WFI01"
    gain_ref["meta"]["instrument"]["name"] = "WFI"
    gain_ref["meta"]["reftype"] = "GAIN"
    gain_ref["meta"]["useafter"] = Time("2022-01-01T11:11:11.111")

    gain_ref["data"] = u.Quantity(
        (rng.random(shape) * 0.5).astype(np.float32) * ingain,
        u.electron / u.DN,
        dtype=np.float32,
    )
    gain_ref["dq"] = np.zeros(shape, dtype=np.uint16)
    gain_ref["err"] = u.Quantity(
        (rng.random(shape) * 0.05).astype(np.float32),
        u.electron / u.DN,
        dtype=np.float32,
    )

    gain_ref_model = GainRefModel(gain_ref)

    # Create temporary readnoise reference file
    rn_ref = maker_utils.mk_readnoise(shape=shape)
    rn_ref["meta"]["instrument"]["detector"] = "WFI01"
    rn_ref["meta"]["instrument"]["name"] = "WFI"
    rn_ref["meta"]["reftype"] = "READNOISE"
    rn_ref["meta"]["useafter"] = Time("2022-01-01T11:11:11.111")

    rn_ref["meta"]["exposure"]["type"] = "WFI_IMAGE"
    rn_ref["meta"]["exposure"]["frame_time"] = 666

    rn_ref["data"] = u.Quantity(
        (rng.random(shape) * 0.01).astype(np.float32), u.DN, dtype=np.float32
    )

    rn_ref_model = ReadnoiseRefModel(rn_ref)

    # return gainfile, readnoisefile
    return gain_ref_model, rn_ref_model



import numpy as np
import pytest
from astropy import coordinates as coord
from astropy import units as u
from astropy.modeling import models
from astropy.table import QTable
from astropy.time import Time
from gwcs import WCS
from gwcs import coordinate_frames as cf
from roman_datamodels import datamodels, maker_utils
from roman_datamodels.maker_utils import mk_common_meta, mk_level2_image

from romancal.datamodels import ModelContainer
from romancal.lib.tests.helpers import word_precision_check
from romancal.resample import gwcs_drizzle, resample_utils
from romancal.resample.resample import (
    ResampleData,
    populate_mosaic_basic,
    populate_mosaic_individual,
)


# Helper function to create a mock input model with specified metadata
def create_mock_model(
    start_time,
    end_time,
    visit,
    segment,
    pass_,
    program,
    survey,
    optical_element,
    instrument_name,
):
    meta = mk_common_meta()
    mock_model = mk_level2_image(**{"meta": meta})
    mock_model.meta.exposure.start_time = Time(start_time, format="mjd")
    mock_model.meta.exposure.end_time = Time(end_time, format="mjd")
    mock_model.meta.exposure.mid_time = Time((start_time + end_time) / 2, format="mjd")
    mock_model.meta.observation.visit = visit
    mock_model.meta.observation.segment = segment
    mock_model.meta.observation["pass"] = pass_
    mock_model.meta.observation.program = program
    mock_model.meta.observation.survey = survey
    mock_model.meta.instrument.optical_element = optical_element
    mock_model.meta.instrument.name = instrument_name
    mock_model.meta.wcsinfo.vparity = -1
    mock_model.meta.wcsinfo.v3yangle = -60
    return mock_model


class WfiSca:
    def __init__(self, fiducial_world, pscale, shape, filename):
        self.fiducial_world = fiducial_world
        self.pscale = pscale
        self.shape = shape
        self.filename = filename

    def create_image(self):
        """
        Create a dummy L2 datamodel given the coordinates of the fiducial point,
        a pixel scale, and the image shape and filename.

        Returns
        -------
        datamodels.ImageModel
            An L2 ImageModel datamodel.
        """
        rng = np.random.default_rng()
        l2 = maker_utils.mk_level2_image(
            shape=self.shape,
            **{
                "meta": {
                    "wcsinfo": {
                        "ra_ref": 10,
                        "dec_ref": 0,
                        "vparity": -1,
                        "v3yangle": -60,
                    },
                    "exposure": {
                        "exposure_time": 152.04000000000002,
                        "effective_exposure_time": 3.04 * 6 * 8,
                    },
                    "observation": {
                        "program": "00005",
                        "execution_plan": 1,
                        "pass": 1,
                        "observation": 1,
                        "segment": 1,
                        "visit": 1,
                        "visit_file_group": 1,
                        "visit_file_sequence": 1,
                        "visit_file_activity": "01",
                        "exposure": 1,
                    },
                },
                "data": u.Quantity(
                    rng.poisson(2.5, size=self.shape).astype(np.float32),
                    u.MJy / u.sr,
                    dtype=np.float32,
                ),
                "var_rnoise": u.Quantity(
                    rng.normal(1, 0.05, size=self.shape).astype(np.float32),
                    u.MJy**2 / u.sr**2,
                    dtype=np.float32,
                ),
                "var_poisson": u.Quantity(
                    rng.poisson(1, size=self.shape).astype(np.float32),
                    u.MJy**2 / u.sr**2,
                    dtype=np.float32,
                ),
                "var_flat": u.Quantity(
                    rng.uniform(0, 1, size=self.shape).astype(np.float32),
                    u.MJy**2 / u.sr**2,
                    dtype=np.float32,
                ),
            },
        )
        # data from WFISim simulation of SCA #01
        l2.meta.filename = self.filename
        l2.meta["wcs"] = create_wcs_object_without_distortion(
            fiducial_world=self.fiducial_world,
            pscale=self.pscale,
            shape=self.shape,
        )
        return datamodels.ImageModel(l2)


def create_wcs_object_without_distortion(fiducial_world, pscale, shape):
    """
    Create a simple WCS object without either distortion or rotation.

    Parameters
    ----------
    fiducial_world : tuple
        A pair of values corresponding to the fiducial's world coordinate.
    pscale : tuple
        A pair of values corresponding to the pixel scale in each axis.
    shape : tuple
        A pair of values specifying the dimensions of the WCS object.

    Returns
    -------
    gwcs.WCS
        A gwcs.WCS object.
    """
    # components of the model
    shift = models.Shift() & models.Shift()

    affine = models.AffineTransformation2D(
        matrix=[[1, 0], [0, 1]], translation=[0, 0], name="pc_rotation_matrix"
    )

    scale = models.Scale(pscale[0]) & models.Scale(pscale[1])

    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(
        fiducial_world[0],
        fiducial_world[1],
        180,
    )

    det2sky = shift | affine | scale | tan | celestial_rotation
    det2sky.name = "linear_transform"

    detector_frame = cf.Frame2D(
        name="detector", axes_names=("x", "y"), unit=(u.pix, u.pix)
    )
    sky_frame = cf.CelestialFrame(
        reference_frame=coord.FK5(), name="fk5", unit=(u.deg, u.deg)
    )

    pipeline = [(detector_frame, det2sky), (sky_frame, None)]

    wcs_obj = WCS(pipeline)

    wcs_obj.bounding_box = (
        (-0.5, shape[-1] - 0.5),
        (-0.5, shape[-2] - 0.5),
    )

    wcs_obj.pixel_shape = shape[::-1]
    wcs_obj.array_shape = shape

    return wcs_obj


@pytest.fixture
def wfi_sca1():
    sca = WfiSca(
        fiducial_world=(10, 0),
        pscale=(0.000031, 0.000031),
        shape=(100, 100),
        filename="r0000501001001001001_01101_0001_WFI01_cal.asdf",
    )

    return sca.create_image()


@pytest.fixture
def wfi_sca2():
    sca = WfiSca(
        fiducial_world=(10.00139, 0),
        pscale=(0.000031, 0.000031),
        shape=(100, 100),
        filename="r0000501001001001001_01101_0001_WFI02_cal.asdf",
    )

    return sca.create_image()


@pytest.fixture
def wfi_sca3():
    sca = WfiSca(
        fiducial_world=(10.00278, 0),
        pscale=(0.000031, 0.000031),
        shape=(100, 100),
        filename="r0000501001001001001_01101_0001_WFI03_cal.asdf",
    )

    return sca.create_image()


@pytest.fixture
def wfi_sca4():
    sca = WfiSca(
        fiducial_world=(10, 0),
        pscale=(0.000031, 0.000031),
        shape=(100, 100),
        filename="r0000501001001001001_01101_0002_WFI01_cal.asdf",
    )

    return sca.create_image()


@pytest.fixture
def wfi_sca5():
    sca = WfiSca(
        fiducial_world=(10.00139, 0),
        pscale=(0.000031, 0.000031),
        shape=(100, 100),
        filename="r0000501001001001001_01101_0002_WFI02_cal.asdf",
    )

    return sca.create_image()


@pytest.fixture
def wfi_sca6():
    sca = WfiSca(
        fiducial_world=(10.00278, 0),
        pscale=(0.000031, 0.000031),
        shape=(100, 100),
        filename="r0000501001001001001_01101_0002_WFI03_cal.asdf",
    )

    return sca.create_image()


@pytest.fixture
def exposure_1(wfi_sca1, wfi_sca2, wfi_sca3):
    """Returns a list with models corresponding to a dummy exposure 1."""
    # set the same exposure time for all SCAs
    for sca in [wfi_sca1, wfi_sca2, wfi_sca3]:
        sca.meta.exposure["start_time"] = Time(
            "2020-02-01T00:00:00", format="isot", scale="utc"
        )
        sca.meta.exposure["end_time"] = Time(
            "2020-02-01T00:02:30", format="isot", scale="utc"
        )
        sca.meta.observation["exposure"] = 1
    return [wfi_sca1, wfi_sca2, wfi_sca3]


@pytest.fixture
def exposure_2(wfi_sca4, wfi_sca5, wfi_sca6):
    """Returns a list with models corresponding to a dummy exposure 2."""
    # set the same exposure time for all SCAs
    for sca in [wfi_sca4, wfi_sca5, wfi_sca6]:
        sca.meta.exposure["start_time"] = Time(
            "2020-05-01T00:00:00", format="isot", scale="utc"
        )
        sca.meta.exposure["end_time"] = Time(
            "2020-05-01T00:02:30", format="isot", scale="utc"
        )
        sca.meta.observation["exposure"] = 2
    return [wfi_sca4, wfi_sca5, wfi_sca6]


@pytest.fixture
def multiple_exposures(exposure_1, exposure_2):
    """Returns a list with all the datamodels from exposure 1 and 2."""
    exposure_1.extend(exposure_2)
    return exposure_1


def test_resampledata_init(exposure_1):
    """Test that ResampleData can set initial values."""
    input_models = exposure_1
    output = "output.asdf"
    single = False
    blendheaders = False
    pixfrac = 0.8
    kernel = "turbo"
    fillval = 0.0
    wht_type = "exp"
    good_bits = "1"
    pscale_ratio = 0.5
    pscale = 0.1
    kwargs = {"in_memory": False}

    resample_data = ResampleData(
        input_models,
        output=output,
        single=single,
        blendheaders=blendheaders,
        pixfrac=pixfrac,
        kernel=kernel,
        fillval=fillval,
        wht_type=wht_type,
        good_bits=good_bits,
        pscale_ratio=pscale_ratio,
        pscale=pscale,
        **kwargs,
    )

    # Assert
    assert resample_data.input_models == input_models
    assert resample_data.output_filename == output
    assert resample_data.pscale_ratio == pscale_ratio
    assert resample_data.single == single
    assert resample_data.blendheaders == blendheaders
    assert resample_data.pixfrac == pixfrac
    assert resample_data.kernel == kernel
    assert resample_data.fillval == fillval
    assert resample_data.weight_type == wht_type
    assert resample_data.good_bits == good_bits
    assert resample_data.in_memory == kwargs["in_memory"]


def test_resampledata_init_default(exposure_1):
    """Test instantiating ResampleData with default values."""
    input_models = exposure_1
    # Default parameter values

    resample_data = ResampleData(input_models)

    # Assert
    assert resample_data.input_models == input_models
    assert resample_data.output_filename is None
    assert resample_data.pscale_ratio == 1.0
    assert not resample_data.single
    assert resample_data.blendheaders
    assert resample_data.pixfrac == 1.0
    assert resample_data.kernel == "square"
    assert resample_data.fillval == "INDEF"
    assert resample_data.weight_type == "ivm"
    assert resample_data.good_bits == "0"
    assert resample_data.in_memory


@pytest.mark.parametrize("input_models", [None, list(), [""], ModelContainer()])
def test_resampledata_init_invalid_input(input_models):
    """Test that ResampleData will raise an exception on invalid inputs."""
    with pytest.raises(Exception) as exec_info:
        ResampleData(input_models)

    assert type(exec_info.value) == ValueError


def test_resampledata_do_drizzle_many_to_one_default_no_rotation_single_exposure(
    exposure_1,
):
    """Test that output WCS encompass the entire combined input WCS region
    by checking that its extrema fall within the output WCS footprint.

    N.B.: since we are not providing the rotation parameter for the
    resample_utils.make_output_wcs method, the output WCS will have
    the same orientation (i.e. same PA) as the detector axes.
    """

    input_models = ModelContainer(exposure_1)
    resample_data = ResampleData(input_models)

    output_models = resample_data.resample_many_to_one()

    output_min_value = np.min(output_models[0].meta.wcs.footprint())
    output_max_value = np.max(output_models[0].meta.wcs.footprint())

    input_wcs_list = [sca.meta.wcs.footprint() for sca in input_models]
    expected_min_value = np.min(np.stack(input_wcs_list))
    expected_max_value = np.max(np.stack(input_wcs_list))

    # Assert
    np.testing.assert_array_less(output_min_value, expected_min_value)
    np.testing.assert_(output_max_value > expected_max_value)


def test_resampledata_do_drizzle_many_to_one_default_no_rotation_multiple_exposures(
    multiple_exposures,
):
    """Test that output WCS encompass the entire combined input WCS region
    by checking that its extrema fall within the output WCS footprint.

    N.B.: since we are not providing the rotation parameter for the
    resample_utils.make_output_wcs method, the output WCS will have
    the same orientation (i.e. same PA) as the detector axes.
    """

    input_models = ModelContainer(multiple_exposures)
    resample_data = ResampleData(input_models)

    output_models = resample_data.resample_many_to_one()

    output_min_value = np.min(output_models[0].meta.wcs.footprint())
    output_max_value = np.max(output_models[0].meta.wcs.footprint())

    input_wcs_list = [sca.meta.wcs.footprint() for sca in multiple_exposures]
    expected_min_value = np.min(np.stack(input_wcs_list))
    expected_max_value = np.max(np.stack(input_wcs_list))

    # Assert
    np.testing.assert_array_less(output_min_value, expected_min_value)
    np.testing.assert_(output_max_value > expected_max_value)


def test_resampledata_do_drizzle_many_to_one_default_rotation_0(exposure_1):
    """Test that output WCS encompass the entire combined input WCS region
    by checking that the output WCS footprint vertices are close to the
    expected vertices for the combined input WCS footprint.

    N.B.: in this case, rotation=0 will create a WCS that will be oriented North up.
    """

    input_models = ModelContainer(exposure_1)
    resample_data = ResampleData(input_models, **{"rotation": 0})

    output_models = resample_data.resample_many_to_one()

    output_min_value = np.min(output_models[0].meta.wcs.footprint())
    output_max_value = np.max(output_models[0].meta.wcs.footprint())

    input_wcs_list = [sca.meta.wcs.footprint() for sca in exposure_1]
    expected_min_value = np.min(np.stack(input_wcs_list))
    expected_max_value = np.max(np.stack(input_wcs_list))

    # Assert
    np.testing.assert_allclose(output_min_value, expected_min_value)
    np.testing.assert_allclose(output_max_value, expected_max_value)


def test_resampledata_do_drizzle_many_to_one_default_rotation_0_multiple_exposures(
    multiple_exposures,
):
    """Test that output WCS encompass the entire combined input WCS region
    by checking that the output WCS footprint vertices are close to the
    expected vertices for the combined input WCS footprint.

    N.B.: in this case, rotation=0 will create a WCS that will be oriented North up.
    """

    input_models = ModelContainer(multiple_exposures)
    resample_data = ResampleData(input_models, **{"rotation": 0})

    output_models = resample_data.resample_many_to_one()

    output_min_value = np.min(output_models[0].meta.wcs.footprint())
    output_max_value = np.max(output_models[0].meta.wcs.footprint())

    input_wcs_list = [sca.meta.wcs.footprint() for sca in multiple_exposures]
    expected_min_value = np.min(np.stack(input_wcs_list))
    expected_max_value = np.max(np.stack(input_wcs_list))

    # Assert
    np.testing.assert_allclose(output_min_value, expected_min_value)
    np.testing.assert_allclose(output_max_value, expected_max_value)


def test_resampledata_do_drizzle_many_to_one_single_input_model(wfi_sca1):
    """Test that the output of resample from a single input file creates a WCS
    footprint vertices that are close to the input WCS footprint's vertices."""

    input_models = ModelContainer([wfi_sca1])
    resample_data = ResampleData(
        input_models, output=wfi_sca1.meta.filename, **{"rotation": 0}
    )

    output_models = resample_data.resample_many_to_one()

    flat_1 = np.sort(wfi_sca1.meta.wcs.footprint().flatten())
    flat_2 = np.sort(output_models[0].meta.wcs.footprint().flatten())

    # Assert
    assert len(output_models) == 1
    assert output_models[0].meta.filename == resample_data.output_filename
    np.testing.assert_allclose(flat_1, flat_2)


def test_update_exposure_times_different_sca_same_exposure(exposure_1):
    """Test that update_exposure_times is properly updating the exposure parameters
    for a set of different SCAs belonging to the same exposure."""
    input_models = ModelContainer(exposure_1)
    resample_data = ResampleData(input_models)

    output_model = resample_data.resample_many_to_one()[0]

    exptime_tot = resample_data.resample_exposure_time(output_model)
    resample_data.update_exposure_times(output_model, exptime_tot)

    # these three SCAs overlap, so the max exposure time is 3x.
    # get this time within 0.1 s.
    time_difference = (
        output_model.meta.resample.product_exposure_time
        - 3 * exposure_1[0].meta.exposure.effective_exposure_time
    )
    assert np.abs(time_difference) < 0.1
    assert (
        output_model.meta.basic.time_first_mjd
        == exposure_1[0].meta.exposure.start_time.mjd
    )
    assert (
        output_model.meta.basic.time_last_mjd
        == exposure_1[0].meta.exposure.end_time.mjd
    )


def test_update_exposure_times_same_sca_different_exposures(exposure_1, exposure_2):
    """Test that update_exposure_times is properly updating the exposure parameters
    for a set of the same SCA but belonging to different exposures."""
    input_models = ModelContainer([exposure_1[0], exposure_2[0]])
    resample_data = ResampleData(input_models)

    output_model = resample_data.resample_many_to_one()[0]

    exptime_tot = resample_data.resample_exposure_time(output_model)
    resample_data.update_exposure_times(output_model, exptime_tot)

    assert len(resample_data.input_models.models_grouped) == 2

    # these exposures overlap perfectly so the max exposure time should
    # be equal to the individual time times two.
    time_difference = (
        output_model.meta.resample.product_exposure_time
        - 2 * exposure_1[0].meta.exposure.effective_exposure_time
    )
    assert np.abs(time_difference) < 0.1

    assert (
        output_model.meta.basic.time_first_mjd
        == min(x.meta.exposure.start_time for x in input_models).mjd
    )

    assert (
        output_model.meta.basic.time_last_mjd
        == max(x.meta.exposure.end_time for x in input_models).mjd
    )

    # likewise the per-pixel median exposure time is just 2x the individual
    # sca exposure time.
    time_difference = (
        output_model.meta.basic.max_exposure_time
        - 2 * exposure_1[0].meta.exposure.effective_exposure_time
    )
    assert np.abs(time_difference) < 0.1


@pytest.mark.parametrize(
    "name",
    ["var_rnoise", "var_poisson", "var_flat"],
)
def test_resample_variance_array(wfi_sca1, wfi_sca4, name):
    """Test that the mean value for the variance array lies within 1% of the
    expectation."""
    input_models = ModelContainer([wfi_sca1, wfi_sca4])
    resample_data = ResampleData(input_models, **{"rotation": 0})

    output_model = resample_data.blank_output.copy()
    output_model.meta["resample"] = {}
    driz = gwcs_drizzle.GWCSDrizzle(
        output_model,
        outwcs=resample_data.output_wcs,
        pixfrac=resample_data.pixfrac,
        kernel=resample_data.kernel,
        fillval=resample_data.fillval,
    )
    [driz.add_image(x.data, x.meta.wcs) for x in resample_data.input_models]

    resample_data.resample_variance_array(name, output_model)

    # combined variance is inversely proportional to the number of "measurements"
    expected_combined_variance_value = np.nanmean(
        [getattr(x, name) for x in input_models]
    ) / len(input_models)

    np.isclose(
        np.nanmean(getattr(output_model, name)).value,
        expected_combined_variance_value,
        atol=0.01,
    )


def test_custom_wcs_input_small_overlap_no_rotation(wfi_sca1, wfi_sca3):
    """Test that resample can create a proper output in the edge case where the
    desired output WCS does not encompass the entire input datamodel but, instead, have
    just a small overlap."""
    input_models = ModelContainer([wfi_sca1])
    resample_data = ResampleData(
        input_models,
        **{"output_wcs": wfi_sca3.meta.wcs, "rotation": 0},
    )

    output_models = resample_data.resample_many_to_one()

    np.testing.assert_allclose(output_models[0].meta.wcs(0, 0), wfi_sca3.meta.wcs(0, 0))


def test_custom_wcs_input_entire_field_no_rotation(multiple_exposures):
    """Test that resample can create a proper output that encompasses the entire
    combined FOV of the input datamodels."""
    input_models = ModelContainer(multiple_exposures)
    # create output WCS encompassing the entire exposure FOV
    output_wcs = resample_utils.make_output_wcs(
        input_models,
        rotation=0,
    )
    resample_data = ResampleData(
        input_models,
        **{"output_wcs": output_wcs},
    )

    output_models = resample_data.resample_many_to_one()

    output_min_value = np.min(output_models[0].meta.wcs.footprint())
    output_max_value = np.max(output_models[0].meta.wcs.footprint())

    input_wcs_list = [sca.meta.wcs.footprint() for sca in multiple_exposures]
    expected_min_value = np.min(np.stack(input_wcs_list))
    expected_max_value = np.max(np.stack(input_wcs_list))

    np.testing.assert_allclose(output_min_value, expected_min_value)
    np.testing.assert_allclose(output_max_value, expected_max_value)


@pytest.mark.parametrize("weight_type", ["ivm", "exptime"])
def test_resampledata_do_drizzle_default_single_exposure_weight_array(
    exposure_1,
    weight_type,
):
    """Test that resample methods return non-empty weight arrays."""

    input_models = ModelContainer(exposure_1)
    resample_data = ResampleData(input_models, wht_type=weight_type)

    output_models_many_to_one = resample_data.resample_many_to_one()
    output_models_many_to_many = resample_data.resample_many_to_many()

    assert np.any(output_models_many_to_one[0].weight > 0)
    assert np.any(output_models_many_to_many[0].weight > 0)


def test_populate_mosaic_basic_single_exposure(exposure_1):
    """
    Test the populate_mosaic_basic function with a given exposure.
    """
    input_models = ModelContainer(exposure_1)
    output_wcs = resample_utils.make_output_wcs(
        input_models,
        pscale_ratio=1,
        pscale=0.000031,
        rotation=0,
        shape=None,
        crpix=(0, 0),
        crval=(0, 0),
    )
    output_model = maker_utils.mk_datamodel(
        datamodels.MosaicModel, shape=tuple(output_wcs.array_shape)
    )

    populate_mosaic_basic(output_model, input_models=input_models)

    input_meta = [datamodel.meta for datamodel in input_models]

    assert output_model.meta.basic.time_first_mjd == np.min(
        [x.exposure.start_time.mjd for x in input_meta]
    )
    assert output_model.meta.basic.time_last_mjd == np.max(
        [x.exposure.end_time.mjd for x in input_meta]
    )
    assert output_model.meta.basic.time_mean_mjd == np.mean(
        [x.exposure.mid_time.mjd for x in input_meta]
    )
    assert output_model.meta.basic.visit == (
        input_meta[0].observation.visit
        if len({x.observation.visit for x in input_meta}) == 1
        else -1
    )
    assert output_model.meta.basic.segment == (
        input_meta[0].observation.segment
        if len({x.observation.segment for x in input_meta}) == 1
        else -1
    )
    assert output_model.meta.basic["pass"] == (
        input_meta[0].observation["pass"]
        if len({x.observation["pass"] for x in input_meta}) == 1
        else -1
    )
    assert output_model.meta.basic.program == (
        input_meta[0].observation.program
        if len({x.observation.program for x in input_meta}) == 1
        else "-1"
    )
    assert output_model.meta.basic.survey == (
        input_meta[0].observation.survey
        if len({x.observation.survey for x in input_meta}) == 1
        else "MULTIPLE"
    )
    assert (
        output_model.meta.basic.optical_element
        == input_meta[0].instrument.optical_element
    )
    assert output_model.meta.basic.instrument == input_meta[0].instrument.name


@pytest.mark.parametrize(
    "input_models_data, expected_output",
    [
        (
            [
                (
                    59000.0,
                    59000.5,
                    1,
                    1,
                    1,
                    "12345",
                    "N/A",
                    "F158",
                    "WFI",
                ),
                (
                    59000.5,
                    59001.0,
                    1,
                    1,
                    1,
                    "12345",
                    "N/A",
                    "F158",
                    "WFI",
                ),
            ],
            {
                "time_first_mjd": 59000.0,
                "time_last_mjd": 59001.0,
                "time_mean_mjd": 59000.5,
                "visit": 1,
                "segment": 1,
                "pass": 1,
                "program": "12345",
                "survey": "N/A",
                "optical_element": "F158",
                "instrument": "WFI",
            },
        ),
        # different visits
        (
            [
                (
                    59000.0,
                    59000.5,
                    1,
                    1,
                    1,
                    "12345",
                    "N/A",
                    "F158",
                    "WFI",
                ),
                (
                    59000.5,
                    59001.0,
                    2,
                    1,
                    1,
                    "12345",
                    "N/A",
                    "F158",
                    "WFI",
                ),
            ],
            {
                "time_first_mjd": 59000.0,
                "time_last_mjd": 59001.0,
                "time_mean_mjd": 59000.5,
                "visit": -1,
                "segment": 1,
                "pass": 1,
                "program": "12345",
                "survey": "N/A",
                "optical_element": "F158",
                "instrument": "WFI",
            },
        ),
        # different segments
        (
            [
                (
                    59000.0,
                    59000.5,
                    1,
                    1,
                    1,
                    "12345",
                    "N/A",
                    "F158",
                    "WFI",
                ),
                (
                    59000.5,
                    59001.0,
                    1,
                    2,
                    1,
                    "12345",
                    "N/A",
                    "F158",
                    "WFI",
                ),
            ],
            {
                "time_first_mjd": 59000.0,
                "time_last_mjd": 59001.0,
                "time_mean_mjd": 59000.5,
                "visit": 1,
                "segment": -1,
                "pass": 1,
                "program": "12345",
                "survey": "N/A",
                "optical_element": "F158",
                "instrument": "WFI",
            },
        ),
        # different passes
        (
            [
                (
                    59000.0,
                    59000.5,
                    1,
                    1,
                    1,
                    "12345",
                    "HLS",
                    "F158",
                    "WFI",
                ),
                (
                    59000.5,
                    59001.0,
                    1,
                    1,
                    2,
                    "12345",
                    "EMS",
                    "F158",
                    "WFI",
                ),
            ],
            {
                "time_first_mjd": 59000.0,
                "time_last_mjd": 59001.0,
                "time_mean_mjd": 59000.5,
                "visit": 1,
                "segment": 1,
                "pass": -1,
                "program": "12345",
                "survey": "MULTIPLE",
                "optical_element": "F158",
                "instrument": "WFI",
            },
        ),
        # different programs
        (
            [
                (
                    59000.0,
                    59000.5,
                    1,
                    1,
                    1,
                    "12345",
                    "N/A",
                    "F158",
                    "WFI",
                ),
                (
                    59000.5,
                    59001.0,
                    1,
                    1,
                    1,
                    "54321",
                    "N/A",
                    "F158",
                    "WFI",
                ),
            ],
            {
                "time_first_mjd": 59000.0,
                "time_last_mjd": 59001.0,
                "time_mean_mjd": 59000.5,
                "visit": 1,
                "segment": 1,
                "pass": 1,
                "program": "-1",
                "survey": "N/A",
                "optical_element": "F158",
                "instrument": "WFI",
            },
        ),
        # different surveys
        (
            [
                (
                    59000.0,
                    59000.5,
                    1,
                    1,
                    1,
                    "12345",
                    "HLS",
                    "F158",
                    "WFI",
                ),
                (
                    59000.5,
                    59001.0,
                    1,
                    1,
                    1,
                    "12345",
                    "EMS",
                    "F158",
                    "WFI",
                ),
            ],
            {
                "time_first_mjd": 59000.0,
                "time_last_mjd": 59001.0,
                "time_mean_mjd": 59000.5,
                "visit": 1,
                "segment": 1,
                "pass": 1,
                "program": "12345",
                "survey": "MULTIPLE",
                "optical_element": "F158",
                "instrument": "WFI",
            },
        ),
    ],
)
def test_populate_mosaic_basic_different_observations(
    input_models_data, expected_output
):
    """Test that populate_mosaic_basic function works properly under different observational scenarios."""
    input_models = [create_mock_model(*data) for data in input_models_data]
    output_wcs = resample_utils.make_output_wcs(
        input_models,
        pscale_ratio=1,
        pscale=0.000031,
        rotation=0,
        shape=None,
        crpix=(0, 0),
        crval=(0, 0),
    )
    output_model = maker_utils.mk_datamodel(
        datamodels.MosaicModel, shape=tuple(output_wcs.array_shape)
    )

    # Act
    populate_mosaic_basic(output_model, input_models)

    # Assert
    assert output_model.meta.basic.time_first_mjd == expected_output["time_first_mjd"]
    assert output_model.meta.basic.time_last_mjd == expected_output["time_last_mjd"]
    assert output_model.meta.basic.time_mean_mjd == expected_output["time_mean_mjd"]
    assert output_model.meta.basic.visit == expected_output["visit"]
    assert output_model.meta.basic.segment == expected_output["segment"]
    assert output_model.meta.basic.program == expected_output["program"]
    assert output_model.meta.basic.survey == expected_output["survey"]
    assert output_model.meta.basic.optical_element == expected_output["optical_element"]
    assert output_model.meta.basic.instrument == expected_output["instrument"]


def test_l3_wcsinfo(multiple_exposures):
    """Test the population of the Level 3 wcsinfo block"""
    expected = maker_utils.mk_mosaic_wcsinfo(
        **{
            "ra_ref": 10.00292450000052,
            "dec_ref": 0.001534500000533253,
            "x_ref": 106.4579605214774,
            "y_ref": 80.66617532540977,
            "rotation_matrix": [
                [-0.9335804264969954, 0.3583679495458379],
                [0.3583679495458379, 0.9335804264969954],
            ],
            "pixel_scale": 3.100000000097307e-05,
            "pixel_scale_local": 3.099999999719185e-05,
            "pixel_shape": (161, 213),
            "ra_center": 10.002930353020417,
            "dec_center": 0.0015101325554100666,
            "ra_corn1": 10.005109345783163,
            "dec_corn1": -0.001982743978690467,
            "ra_corn2": 10.006897960220385,
            "dec_corn2": 0.002676755917536623,
            "ra_corn3": 10.000733528663718,
            "dec_corn3": 0.005043059486913547,
            "ra_corn4": 9.998944914237953,
            "dec_corn4": 0.00038355958555111314,
            "orientat_local": 9.826983262839223,
            "orientat": 9.826978421513601,
            "projection": "TAN",
            "s_region": (
                "POLYGON ICRS 10.005109345783163 -0.001982743978690467 10.006897960220385 "
                "0.002676755917536623 10.000733528663718 0.005043059486913547 "
                "9.998944914237953 0.00038355958555111314 "
            ),
        }
    )

    input_models = ModelContainer(multiple_exposures)
    resample_data = ResampleData(input_models)

    output_model = resample_data.resample_many_to_one()[0]

    assert output_model.meta.wcsinfo.projection == expected.projection
    assert word_precision_check(output_model.meta.wcsinfo.s_region, expected.s_region)
    for key in expected.keys():
        if key not in ["projection", "s_region"]:
            assert np.allclose(output_model.meta.wcsinfo[key], expected[key])


def test_l3_individual_image_meta(multiple_exposures):
    """Test that the individual_image_meta is being populated"""
    input_models = ModelContainer(multiple_exposures)
    output_model = maker_utils.mk_datamodel(datamodels.MosaicModel)

    # Act
    populate_mosaic_individual(output_model, input_models)

    # Assert sizes are expected
    n_inputs = len(input_models)
    for value in output_model.meta.individual_image_meta.values():
        assert isinstance(value, QTable)
        assert len(value) == n_inputs

    # Assert spot check on filename, which is different for each mock input
    basic_table = output_model.meta.individual_image_meta.basic
    for idx, input in enumerate(input_models):
        assert input.meta.filename == basic_table["filename"][idx]



"""Test basic utils"""

import numpy as np
import pytest
from astropy.table import Table

from romancal.lib.basic_utils import bytes2human, recarray_to_ndarray

test_data = [
    (1000, "1000B"),
    (1024, "1.0K"),
    (1024 * 10, "10.0K"),
    (100001221, "95.4M"),
]


@pytest.mark.parametrize("input_data, result", test_data)
def test_bytes2human(input_data, result):
    """test the basic conversion"""

    assert bytes2human(input_data) == result


def test_structured_array_utils():
    arrays = [np.arange(0, 10), np.arange(10, 20), np.arange(30, 40)]
    names = "a, b, c"

    recarr0 = np.core.records.fromarrays(
        arrays, names=names, formats=[arr.dtype for arr in arrays]
    )
    round_tripped = recarray_to_ndarray(recarr0)
    assert np.all(round_tripped == np.column_stack(arrays).astype("<f8"))

    astropy_table = Table(recarr0)
    assert np.all(
        np.array(arrays) == np.array([col.data for col in astropy_table.itercols()])
    )

    assert np.all(astropy_table.as_array() == recarr0)



""" Roman tests for flat field correction """

import pytest
import roman_datamodels as rdm
from metrics_logger.decorators import metrics_logger

from romancal.pipeline.exposure_pipeline import ExposurePipeline


def passfail(bool_expr):
    """set pass fail"""
    if bool_expr:
        return "Pass"
    return "Fail"


@pytest.mark.bigdata
@pytest.mark.soctests
@metrics_logger("DMS413")
def test_16resultants_image_processing(rtdata, ignore_asdf_paths):
    """Tests for imaging processing requirements for 16 resultants (DMS413)"""
    # The input data is from INS for stress testing at some point this should be generated
    # every time new data is needed.

    input_data = "r0000101001001001001_01101_0004_WFI01_uncal.asdf"
    rtdata.get_data(f"WFI/image/{input_data}")
    rtdata.input = input_data

    # Test Pipeline
    output = "r0000101001001001001_01101_0004_WFI01_cal.asdf"
    rtdata.output = output
    args = [
        "--disable-crds-steppars",
        "roman_elp",
        rtdata.input,
    ]
    ExposurePipeline.from_cmdline(args)

    # Perform DMS tests
    # Initial prep
    pipeline = ExposurePipeline()
    # rtdata.get_data(f"WFI/image/{output}")
    model = rdm.open(output, lazy_load=False)
    uncal_data = rdm.open(input_data, lazy_load=False)

    # DMS280 result is an ImageModel
    pipeline.log.info(
        "DMS413 MSG: Testing that result is a Level 2 model......."
        + passfail(isinstance(model, rdm.datamodels.ImageModel))
    )

    pipeline.log.info(
        "DMS413 MSG: Testing that there are 16 resultants in the input file......."
        + passfail(len(uncal_data.meta.exposure.read_pattern) == 16)
    )

    pipeline.log.info(
        "DMS413 MSG: Testing that there are 16 resultants listed in the output file......."
        + passfail(len(model.meta.exposure.read_pattern) == 16)
    )

    # Ensure step completion is as expected
    assert model.meta.cal_step.dq_init == "COMPLETE"
    assert model.meta.cal_step.saturation == "COMPLETE"
    assert model.meta.cal_step.linearity == "COMPLETE"
    assert model.meta.cal_step.dark == "COMPLETE"
    assert model.meta.cal_step.jump == "COMPLETE"
    assert model.meta.cal_step.ramp_fit == "COMPLETE"
    assert model.meta.cal_step.assign_wcs == "COMPLETE"
    assert model.meta.cal_step.flat_field == "COMPLETE"
    assert model.meta.cal_step.photom == "COMPLETE"


@pytest.mark.bigdata
@pytest.mark.soctests
@metrics_logger("DMS414")
def test_16resultants_spectral_processing(rtdata, ignore_asdf_paths):
    """Tests for imaging processing requirements for 16 resultants (DMS413)"""
    # The input data is from INS for stress testing at some point this should be generated
    # by INS every time new data is needed.

    input_data = "r0000201001001001001_01101_0004_WFI01_uncal.asdf"
    rtdata.get_data(f"WFI/grism/{input_data}")
    rtdata.input = input_data

    # Test Pipeline
    output = "r0000201001001001001_01101_0004_WFI01_cal.asdf"
    rtdata.output = output
    args = [
        "--disable-crds-steppars",
        "roman_elp",
        rtdata.input,
    ]
    ExposurePipeline.from_cmdline(args)

    # Perform DMS tests
    # Initial prep
    pipeline = ExposurePipeline()
    # rtdata.get_data(f"WFI/image/{output}")
    model = rdm.open(output, lazy_load=False)
    uncal_data = rdm.open(input_data, lazy_load=False)

    # DMS280 result is an ImageModel
    pipeline.log.info(
        "DMS414 MSG: Testing that result is a Level 2 model......."
        + passfail(isinstance(model, rdm.datamodels.ImageModel))
    )

    pipeline.log.info(
        "DMS414 MSG: Testing that there are 16 resultants in the input file......."
        + passfail(len(uncal_data.meta.exposure.read_pattern) == 16)
    )

    pipeline.log.info(
        "DMS414 MSG: Testing that there are 16 resultants listed in the output file......."
        + passfail(len(model.meta.exposure.read_pattern) == 16)
    )

    # Ensure step completion is as expected
    assert model.meta.cal_step.dq_init == "COMPLETE"
    assert model.meta.cal_step.saturation == "COMPLETE"
    assert model.meta.cal_step.linearity == "COMPLETE"
    assert model.meta.cal_step.dark == "COMPLETE"
    assert model.meta.cal_step.jump == "COMPLETE"
    assert model.meta.cal_step.ramp_fit == "COMPLETE"
    assert model.meta.cal_step.assign_wcs == "COMPLETE"
    assert model.meta.cal_step.flat_field == "SKIPPED"
    assert model.meta.cal_step.photom == "SKIPPED"



#  Module for 2d saturation
#
import logging

import numpy as np
from roman_datamodels.dqflags import pixel
from stcal.saturation.saturation import flag_saturated_pixels

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

ATOD_LIMIT = 65535.0  # Hard DN limit of 16-bit A-to-D converter


def flag_saturation(input_model, ref_model):
    """
    Short Summary
    -------------
    Function to call stcal for flagging saturated pixels.

    Parameters
    ----------
    input_model : `~roman_datamodels.datamodels.RampModel`
        The input science data to be corrected

    ref_model : `~roman_datamodels.datamodels.SaturationRefModel`
        Saturation reference file data model

    Returns
    -------
    output_model : `~roman_datamodels.datamodels.RampModel`
        Data model with saturation, A/D floor, and do not use flags set in
        the GROUPDQ array
        The input model is modified in place and returned as the output model.
    """

    data = input_model.data[np.newaxis, :].value

    # Modify input_model in place.
    gdq = input_model.groupdq[np.newaxis, :]
    pdq = input_model.pixeldq[np.newaxis, :]

    # Copy information from saturation reference file
    sat_thresh = ref_model.data.value.copy()
    sat_dq = ref_model.dq.copy()

    # Obtain dq arrays updated for saturation
    # The third variable is the processed ZEROFRAME, which is not
    # used in romancal, so is always None.
    gdq_new, pdq_new, _ = flag_saturated_pixels(
        data,
        gdq,
        pdq,
        sat_thresh,
        sat_dq,
        ATOD_LIMIT,
        pixel,
        n_pix_grow_sat=0,
        read_pattern=input_model.meta.exposure.read_pattern,
    )

    # Save the flags in the output GROUPDQ array
    input_model.groupdq = gdq_new[0, :]

    # Save the NO_SAT_CHECK flags in the output PIXELDQ array
    input_model.pixeldq = pdq_new[0, :]

    return input_model



"""Key/Value Registry"""

from collections import UserDict

__all__ = [
    "KeyValueRegistry",
    "KeyValueRegistryError",
    "KeyValueRegistryNoKeyFound",
    "KeyValueRegistryNotSingleItemError",
]


class KeyValueRegistry(UserDict):
    """Provide a dict-like registry

    Differences from just a `dict`:
        - Can be given single item or a 2-tuple.
          If an item, attempts to read the `__name__` attribute
          and use that as the key.

        - If None is given as a key, a default key can
          be specified.

        - Instances can be used as decorators.

    Parameters
    ----------
    items : object or (str, object) or dict
        Initializing items.

    default : str or object
        The default to use when key is `None`
    """

    def __init__(self, items=None, default=None):
        super_args = ()
        if items is not None:
            super_args = (make_dict(items),)
        super().__init__(*super_args)

        self.default = None
        if default is not None:
            default_dict = make_dict(default)
            if len(default_dict) > 1:
                raise KeyValueRegistryNotSingleItemError
            default_dict = make_dict(default)
            self.update(default_dict)
            self.default = next(iter(default_dict.keys()))
            self.update({None: default_dict[self.default]})

    def update(self, item):
        """Add item to registry"""
        item_dict = make_dict(item)
        super().update(item_dict)

    def __call__(self, item):
        """Add item by calling instance

        This allows an instance to be used as a decorator.
        """
        self.update(item)
        return item


# ******
# Errors
# ******
class KeyValueRegistryError(Exception):
    def __init__(self, *args):
        if len(args) == 0:
            args = (self.msg,)
        super().__init__(*args)


class KeyValueRegistryNotSingleItemError(KeyValueRegistryError):
    msg = "Item cannot be a list"


class KeyValueRegistryNoKeyFound(KeyValueRegistryError):
    msg = "Cannot deduce key from given value"


# *********
# Utilities
# *********
def make_dict(item):
    """Create a dict from an item

    Parameters
    ----------
    item : object or (name, object) or dict
        If dict, just return dict.
        If 2-tuple, return dict with the key/value pair
        If just object, use `__name__` as key
    """
    try:
        item_dict = dict(item)
    except (TypeError, ValueError):
        try:
            key, value = item
        except (TypeError, ValueError):
            try:
                key = item.__name__
            except (AttributeError, SyntaxError):
                raise KeyValueRegistryNoKeyFound
            else:
                value = item

        # At they point we have a key/value pair
        item_dict = {key: value}

    return item_dict



"""General utility objects"""

import numpy as np
from roman_datamodels.datamodels import AssociationsModel
from roman_datamodels.dqflags import group, pixel


def bytes2human(n):
    """Convert bytes to human-readable format

    Taken from the `psutil` library which references
    http://code.activestate.com/recipes/578019

    Parameters
    ----------
    n : int
        Number to convert

    Returns
    -------
    readable : str
        A string with units attached.

    Examples
    --------
    >>> bytes2human(10000)
        '9.8K'

    >>> bytes2human(100001221)
        '95.4M'
    """
    symbols = ("K", "M", "G", "T", "P", "E", "Z", "Y")
    prefix = {}
    for i, s in enumerate(symbols):
        prefix[s] = 1 << (i + 1) * 10
    for s in reversed(symbols):
        if n >= prefix[s]:
            value = float(n) / prefix[s]
            return f"{value:.1f}{s}"
    return f"{n}B"


def is_fully_saturated(model):
    """
    Check to see if all data pixels are flagged as saturated.
    """

    if np.all(np.bitwise_and(model.groupdq, group.SATURATED) == group.SATURATED):
        return True
    elif np.all(np.bitwise_and(model.pixeldq, pixel.SATURATED) == pixel.SATURATED):
        return True

    return False


def is_association(asn_data):
    """
    Test if an object is an association by checking for required fields
    """
    return AssociationsModel.is_association(asn_data)


class LoggingContext:
    """Logging context manager
    Keep logging configuration within a context
    Based on the Python 3 Logging Cookbook example
    Parameters
    ==========
    logger: logging.Logger
        The logger to modify.
    level: int
        The log level to set.
    handler: logging.Handler
        The handler to use.
    close: bool
        Close the handler when done.
    """

    def __init__(self, logger, level=None, handler=None, close=True):
        self.logger = logger
        self.level = level
        self.handler = handler
        self.close = close

        self.old_level = None

    def __enter__(self):
        if self.level is not None:
            self.old_level = self.logger.level
            self.logger.setLevel(self.level)
        if self.handler:
            self.logger.addHandler(self.handler)

    def __exit__(self, et, ev, tb):
        if self.level is not None:
            self.logger.setLevel(self.old_level)
        if self.handler:
            self.logger.removeHandler(self.handler)
        if self.handler and self.close:
            self.handler.close()
        # implicit return of None => don't swallow exceptions


def recarray_to_ndarray(x, to_dtype="<f8"):
    """
    Convert a structured array to a 2D ndarray.

    Parameters
    ----------
    x : np.record
        Structured array
    to_dtype : str
        Cast all columns in `x` to this dtype in the output ndarray.

    Returns
    -------
    array : np.ndarray
        Numpy array (without labeled columns).
    """
    names = x.dtype.names
    astype = [(name, to_dtype) for name in names]
    return np.asarray(x.astype(astype).view(to_dtype).reshape((-1, len(names))))



import os
from contextlib import contextmanager

import pytest

# chdir contextmanager was added with Python 3.11
try:
    from contextlib import chdir
except ImportError:

    @contextmanager
    def chdir(path):
        old_path = os.getcwd()

        try:
            os.chdir(path)
            yield
        finally:
            os.chdir(old_path)


@pytest.fixture(autouse=True)
def _docdir(request):
    # Trigger ONLY for doctestplus.
    doctest_plugin = request.config.pluginmanager.getplugin("doctestplus")
    if isinstance(request.node.parent, doctest_plugin._doctest_textfile_item_cls):
        tmp_path = request.getfixturevalue("tmp_path")
        with chdir(tmp_path):
            yield
    else:
        yield



#! /usr/bin/env python

from roman_datamodels import datamodels as rdm

from romancal.stpipe import RomanStep

__all__ = ["DarkCurrentStep"]


class DarkCurrentStep(RomanStep):
    """
    DarkCurrentStep: Performs dark current correction by subtracting
    dark current reference data from the input science data model.
    """

    spec = """
        dark_output = output_file(default = None) # Dark corrected model
    """

    reference_file_types = ["dark"]

    def process(self, input):
        # Open the input data model
        with rdm.open(input, lazy_load=False) as input_model:
            # Get the name of the dark reference file to use
            self.dark_name = self.get_reference_file(input_model, "dark")
            # Check for a valid reference file
            if self.dark_name == "N/A":
                self.log.warning("No DARK reference file found")
                self.log.warning("Dark current step will be skipped")
                result = input_model
                result.meta.cal_step.dark = "SKIPPED"
                return result

            self.log.info("Using DARK reference file: %s", self.dark_name)

            # Open dark model
            dark_model = rdm.open(self.dark_name)

            # Temporary patch to utilize stcal dark step until MA table support
            # is fully implemented
            if "ngroups" not in dark_model.meta.exposure:
                dark_model.meta.exposure["ngroups"] = dark_model.data.shape[0]
            if "nframes" not in dark_model.meta.exposure:
                dark_model.meta.exposure["nframes"] = input_model.meta.exposure.nframes
            if "groupgap" not in dark_model.meta.exposure:
                dark_model.meta.exposure["groupgap"] = (
                    input_model.meta.exposure.groupgap
                )

            # Do the dark correction
            out_model = input_model
            nresultants = len(input_model.meta.exposure["read_pattern"])
            out_model.data -= dark_model.data[:nresultants]
            out_model.pixeldq |= dark_model.dq
            out_model.meta.cal_step.dark = "COMPLETE"

            # Save dark data to file
            if self.dark_output is not None:
                dark_model.save(self.dark_output)
                # not clear to me that this makes any sense for Roman
            dark_model.close()

        if self.save_results:
            try:
                self.suffix = "darkcurrent"
            except AttributeError:
                self["suffix"] = "darkcurrent"
        return out_model



"""Test versioning consistency"""

from romancal import __version__
from romancal.associations.asn_from_list import asn_from_list


def test_asn_version():
    """Test version in association is package version"""

    asn = asn_from_list(["a", "b", "c"], product_name="aproduct")

    assert asn["code_version"] == __version__



#!/usr/bin/env python
"""
Regression test okifying script.

Requires JFrog CLI (https://jfrog.com/getcli/) configured with credentials
that have write access to the roman-pipeline repository.
"""

import os
import shutil
import subprocess
import tempfile
from argparse import ArgumentParser
from contextlib import contextmanager
from glob import glob

import asdf

ARTIFACTORY_REPO = "roman-pipeline-results"
SPECFILE_SUFFIX = "_okify.json"
RTDATA_SUFFIX = "_rtdata.asdf"
TERMINAL_WIDTH = shutil.get_terminal_size((80, 20)).columns


def parse_args():
    parser = ArgumentParser(description="Okify regression test results")
    parser.add_argument(
        "build_number",
        help="Jenkins build number for Roman builds",
        metavar="build-number",
    )
    parser.add_argument(
        "--job-name",
        help="Jenkins job name under [RT] (default: romancal)",
        default="romancal",
        metavar="job-name",
    )
    parser.add_argument(
        "--dry-run", action="store_true", help="pass the --dry-run flag to JFrog CLI"
    )

    return parser.parse_args()


def artifactory_copy(specfile, dry_run=False):
    jfrog_args = []

    if dry_run:
        jfrog_args.append("--dry-run")

    args = list(["jfrog", "rt", "cp"] + jfrog_args + [f"--spec={specfile}"])
    subprocess.run(args, check=True)


def artifactory_get_breadcrumbs(build_number, job_name, suffix):
    """Download specfiles or other breadcrump from Artifactory associated with
    a build number and return a list of their locations on the local file system

    An example search would be:

    jfrog rt search roman-pipeline-results/*/*_okify.json --props='build.number=540;build.name=RT :: romancal'
    """  # noqa: E501
    build_name = f"{job_name}"

    # Retreive all the okify specfiles for failed tests.
    args = list(
        ["jfrog", "rt", "dl"]
        + [f"{ARTIFACTORY_REPO}/*RT-{build_name}-{build_number}*/*{suffix}"]
    )
    subprocess.run(args, check=True, capture_output=True)

    return sorted(glob(f"**/**/*{suffix}"))


def artifactory_get_build_artifacts(build_number, job_name):
    specfiles = artifactory_get_breadcrumbs(build_number, job_name, SPECFILE_SUFFIX)
    asdffiles = artifactory_get_breadcrumbs(build_number, job_name, RTDATA_SUFFIX)

    if len(specfiles) != len(asdffiles):
        raise RuntimeError("Different number of _okify.json and _rtdata.asdf files")

    for a, b in zip(specfiles, asdffiles):
        if a.replace(SPECFILE_SUFFIX, "") != b.replace(RTDATA_SUFFIX, ""):
            raise RuntimeError("The _okify.json and _rtdata.asdf files are not matched")

    return specfiles, asdffiles


@contextmanager
def pushd(newdir):
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)


def main():
    args = parse_args()

    build = args.build_number
    name = args.job_name

    # Create and chdir to a temporary directory to store specfiles
    with tempfile.TemporaryDirectory() as tmpdir:
        print(f"Downloading test okify artifacts to local directory {tmpdir}")
        with pushd(tmpdir):
            # Retreive all the okify specfiles for failed tests.
            specfiles, asdffiles = artifactory_get_build_artifacts(build, name)

            number_failed_tests = len(specfiles)

            print(f"{number_failed_tests} failed tests to okify")

            for i, (specfile, asdffile) in enumerate(zip(specfiles, asdffiles)):
                # Print traceback and OKify info for this test failure
                with asdf.open(asdffile) as af:
                    traceback = af.tree["traceback"]
                    remote_results_path = af.tree["remote_results_path"]
                    output = af.tree["output"]
                    truth_remote = af.tree["truth_remote"]
                    try:
                        test_name = af.tree["test_name"]
                    except KeyError:
                        test_name = "test_name"

                remote_results = os.path.join(
                    remote_results_path, os.path.basename(output)
                )

                test_number = i + 1

                print(f" {test_name} ".center(TERMINAL_WIDTH, "—"))
                print(traceback)
                print("—" * TERMINAL_WIDTH)
                print(f"OK: {remote_results}")
                print(f"--> {truth_remote}")
                print(
                    f"[ test {test_number} of {number_failed_tests} ]".center(
                        TERMINAL_WIDTH, "—"
                    )
                )

                # Ask if user wants to okify this test
                while True:
                    result = input("Enter 'o' to okify, 's' to skip: ")
                    if result not in ["o", "s"]:
                        print(f"Unrecognized command '{result}', try again")
                    else:
                        break

                if result == "s":
                    print("Skipping\n")
                else:
                    artifactory_copy(os.path.abspath(specfile), dry_run=args.dry_run)
                    print("")


if __name__ == "__main__":
    main()



import logging

import roman_datamodels as rdm

from romancal.refpix import refpix
from romancal.stpipe import RomanStep

__all__ = ["RefPixStep"]

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class RefPixStep(RomanStep):
    """
    RefpixStep: Module for correcting the the science data using the reference
        pixels
    """

    spec = """
    remove_offset = boolean(default=True) # Turn on or off removing the data offset
    # prior to the reference pixel correction, then returning the offset afterwords.
    remove_trends = boolean(default=True) # Turn on or off removing the boundary
    # linear trends
    cosine_interpolate = boolean(default=True) # Turn on or off the cosine
    # interpolation of the reference pixels
    fft_interpolate = boolean(default=True) # Turn on or off the FFT interpolation
    # of the reference pixel padded values.
    """

    reference_file_types = ["refpix"]

    def process(self, input):
        """
        Perform the reference pixel correction
        """

        # open the input data model
        log.debug(f"Opening the science data: {input}")
        datamodel = rdm.open(input)

        # Get the reference file
        ref_file = self.get_reference_file(datamodel, "refpix")

        # Check for a valid reference file
        if ref_file == "N/A":
            self.log.warning("No REFPIX reference file found")
            self.log.warning("Reference pixel correction step will be skipped")
            result = datamodel
            result.meta.cal_step.refpix = "SKIPPED"
            return result

        log.debug(f"Opening the reference file: {ref_file}")
        with rdm.open(ref_file) as refs:
            # Run the correction
            log.debug("Running the reference pixel correction")
            output = refpix.run_steps(
                datamodel,
                refs,
                self.remove_offset,
                self.remove_trends,
                self.cosine_interpolate,
                self.fft_interpolate,
            )
            # Update the step status
            output.meta.cal_step["refpix"] = "COMPLETE"
            if self.save_results:
                try:
                    self.suffix = "refpix"
                except AttributeError:
                    self["suffix"] = "refpix"
        return output



"""
Module for applying flat fielding
"""

import logging

import numpy as np
from astropy import units as u
from roman_datamodels.dqflags import pixel

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

MICRONS_100 = 1.0e-4  # 100 microns, in meters


def do_correction(input_model, flat=None):
    """Flat-field a Roman data model using a flat-field model

    Parameters
    ----------
    input_model : Roman data model
        Input science data model to be flat-fielded.

    flat : Roman data model, or None
        Data model containing flat-field for all instruments

    Returns
    -------
    output_model : data model
        The data model for the flat-fielded science data.
        The data is modified in place.
    """

    do_flat_field(input_model, flat)

    return input_model


def do_flat_field(output_model, flat_model):
    """Apply flat-fielding, and update the output model.

    Parameters
    ----------
    output_model : Roman data model
        flat-fielded input science data model, modified in-place

    flat_model : Roman data model
        data model containing flat-field
    """
    if flat_model is not None and output_model.data.shape != flat_model.data.shape:
        # Check to see if flat data array is smaller than science data
        log.warning("Flat data array is not the same shape as the science data")
        log.warning("Step will be skipped")
        output_model.meta.cal_step.flat_field = "SKIPPED"
    elif output_model.meta.exposure.type != "WFI_IMAGE":
        # Check to see if attempt to flatten non-Image data
        log.info("Skipping flat field for spectral exposure.")
        output_model.meta.cal_step.flat_field = "SKIPPED"
    elif flat_model is None:
        # Check to see if attempt to flatten non-Image data
        log.info("Skipping flat field - no flat reference file.")
        output_model.meta.cal_step.flat_field = "SKIPPED"
    else:
        apply_flat_field(output_model, flat_model)
        output_model.meta.cal_step.flat_field = "COMPLETE"


def apply_flat_field(science, flat):
    """Flat field the data and error arrays.

    Extended summary
    ----------------
    The science data and error arrays will be divided by the flat field.
    The data quality array will be updated based on bad pixels in flat
    field arrays. Applies portion of flat field corresponding to science
    image subarray.

    Parameters
    ----------
    science : Roman data model
        input science data model

    flat : Roman data model
        flat field data model
    """
    flat_data = flat.data.copy()
    flat_dq = flat.dq.copy()
    flat_err = flat.err.copy()
    # Find pixels in the flat that have a value of NaN and set
    # their DQ to NO_FLAT_FIELD
    flat_nan = np.isnan(flat_data)
    flat_dq[flat_nan] = np.bitwise_or(flat_dq[flat_nan], pixel.NO_FLAT_FIELD)

    # Find pixels in the flat that have a value of zero, and set
    # their DQ to NO_FLAT_FIELD
    flat_zero = np.where(flat_data == 0.0)
    flat_dq[flat_zero] = np.bitwise_or(flat_dq[flat_zero], pixel.NO_FLAT_FIELD)

    # Find all pixels in the flat that have a DQ value of NO_FLAT_FIELD
    flat_bad = np.bitwise_and(flat_dq, pixel.NO_FLAT_FIELD)

    # Reset the flat value of all bad pixels to 1.0, so that no
    # correction is made
    flat_data[np.where(flat_bad)] = 1.0
    # Now let's apply the correction to science data and error arrays.  Rely
    # on array broadcasting to handle the cubes
    science.data = u.Quantity(
        (science.data.value / flat_data), u.DN / u.s, dtype=science.data.dtype
    )

    # Update the variances using BASELINE algorithm.  For guider data, it has
    # not gone through ramp fitting so there is no Poisson noise or readnoise
    flat_data_squared = flat_data**2
    science.var_poisson /= flat_data_squared
    science.var_rnoise /= flat_data_squared
    try:
        science.var_flat = science.data**2 / flat_data_squared * flat_err**2
    except AttributeError:
        science["var_flat"] = np.zeros(shape=science.data.shape, dtype=np.float32)
        science.var_flat = science.data**2 / flat_data_squared * flat_err**2

    science.err = np.sqrt(science.var_poisson + science.var_rnoise + science.var_flat)
    science.err = science.err.to(science.data.unit)
    # Workaround for https://github.com/astropy/astropy/issues/16055

    # Combine the science and flat DQ arrays
    science.dq = np.bitwise_or(science.dq, flat_dq)



"""Treat various objects as Associations"""

from functools import partial
from os import path as os_path

from ..associations import Association, AssociationRegistry, libpath, load_asn
from ..associations.asn_from_list import asn_from_list
from ..associations.lib.rules_elpp_base import DMS_ELPP_Base
from ..associations.lib.rules_level2 import Asn_Lv2Image

__all__ = [
    "LoadAsAssociation",
    "LoadAsLevel2Asn",
]


DEFAULT_NAME = "singleton"
DEFAULT_ASN_META = {
    "program": DEFAULT_NAME,
    "target": DEFAULT_NAME,
    "asn_pool": DEFAULT_NAME,
}


class LoadAsAssociation(dict):
    """Read in or create an association

    Parameters
    ----------
    asn : dict or Association
        An already existing association

    Notes
    -----
    This class is normally not instantiated.
    the `load` method should be used as the factory
    method to read an association or create one from
    a string or `Datamodel` object, or a list of such
    objects.
    """

    @classmethod
    def load(
        cls,
        obj,
        meta=DEFAULT_ASN_META,
        registry=AssociationRegistry,
        rule=Association,
        product_name_func=None,
    ):
        """Load object and return an association of it

        Parameters
        ----------
        obj : Association, str, Datamodel, [str[,...]], [Datamodel[,...]]
            The obj to return as an association

        registry : AssociationRegistry
            The registry to use to load an association file with

        rule : Association
            The rule to use if an association needs to be created

        product_name_func : func
            A function, when given the argument of `obj`, or
            if `obj` is a list, each item in `obj`, returns
            a string that will be used as the product name in
            the association.

        Attributes
        ----------
        Along with the attributes belonging to an association, the
        following are added:

        filename : str
            The name of the association file, if such a file
            where passed in. Otherwise a default value is given.

        Returns
        -------
        association : Association
            An association created using given obj
        """
        try:
            with open(obj) as fp:
                pure_asn = load_asn(fp, registry=registry)
        except Exception:
            if not isinstance(obj, list):
                obj = [obj]
            asn = asn_from_list(
                obj,
                rule=rule,
                meta=DEFAULT_ASN_META,
                product_name_func=product_name_func,
            )
            asn.filename = DEFAULT_NAME
        else:
            asn = rule()
            asn.update(pure_asn)
            asn.filename = obj

        return asn


class LoadAsLevel2Asn(LoadAsAssociation):
    """Read in or create a Level2 association"""

    @classmethod
    def load(cls, obj, basename=None):
        """Open object and return a Level2 association of it

        Parameters
        ----------
        obj : Association, str, Datamodel, [str[,...]], [Datamodel[,...]]
            The obj to return as an association

        basename : str
            If specified, use as the basename, with an index appended.

        Attributes
        ----------
        Along with the attributes belonging to a Level2 association, the
        following are added:

        filename : str
            The name of the association file, if such a file
            where passed in. Otherwise a default value is given.

        Returns
        -------
        association : DMSLevel2bBase
            An association created using given obj
        """
        product_name_func = cls.model_product_name
        if basename is not None:
            product_name_func = partial(cls.name_with_index, basename)

        # if the input string is a FITS file create an asn and return
        if isinstance(obj, str):
            file_name, file_ext = os_path.splitext(obj)

            if file_ext == ".fits":
                items = [(obj, "science")]
                asn = asn_from_list(
                    items,
                    product_name=file_name,
                    rule=DMS_ELPP_Base,
                    with_exptype=True,
                    meta={"asn_pool": "singleton"},
                )
                return asn

        asn = super().load(
            obj,
            registry=AssociationRegistry(
                definition_files=[libpath("rules_level2.py")], include_default=False
            ),
            rule=Asn_Lv2Image,
            product_name_func=product_name_func,
        )
        return asn

    @staticmethod
    def model_product_name(model, *args, **kwargs):
        """Product a model product name based on the model.

        Parameters
        ----------
        model : DataModel
            The model to get the name from

        Returns
        -------
        product_name : str
            The basename of filename from the model
        """
        product_name, ext = os_path.splitext(model.meta.filename)
        return product_name

    @staticmethod
    def name_with_index(basename, obj, idx, *args, **kwargs):
        """Produce a name with the basename and index appended

        Parameters
        ----------
        basename : str
            The base of the file name

        obj : object
            The object being added to the association _(unused)_

        idx : int
            The current index of the added item.

        Returns
        -------
        product_name : str
            The concatenation of basename, '_', idx

        Notes
        -----
        If the index is less than or equal to 1, no appending is done.
        """
        basename, extension = os_path.splitext(os_path.basename(basename))
        if idx > 1:
            basename = basename + "_" + str(idx)
        return basename



#!/usr/bin/env python
import logging
from os.path import basename

import numpy as np
from roman_datamodels import datamodels as rdm
from roman_datamodels.dqflags import group

import romancal.datamodels.filetype as filetype

# step imports
from romancal.assign_wcs import AssignWcsStep

# from romancal.associations.exceptions import AssociationNotValidError
# from romancal.associations.load_as_asn import LoadAsLevel2Asn
from romancal.dark_current import DarkCurrentStep
from romancal.datamodels import ModelContainer
from romancal.dq_init import dq_init_step
from romancal.flatfield import FlatFieldStep
from romancal.lib.basic_utils import is_fully_saturated
from romancal.linearity import LinearityStep
from romancal.photom import PhotomStep
from romancal.ramp_fitting import ramp_fit_step
from romancal.refpix import RefPixStep
from romancal.saturation import SaturationStep
from romancal.source_detection import SourceDetectionStep
from romancal.tweakreg import TweakRegStep

from ..stpipe import RomanPipeline

__all__ = ["ExposurePipeline"]

# Define logging
log = logging.getLogger()
log.setLevel(logging.DEBUG)


class ExposurePipeline(RomanPipeline):
    """
    ExposurePipeline: Apply all calibration steps to raw Roman WFI
    ramps to produce a 2-D slope product. Included steps are:
    dq_init, saturation, linearity, dark current, jump detection, ramp_fit,
    assign_wcs, flatfield (only applied to WFI imaging data), photom,
    and source_detection.
    """

    class_alias = "roman_elp"

    spec = """
        save_calibrated_ramp = boolean(default=False)
        save_results = boolean(default=False)
    """

    # Define aliases to steps
    step_defs = {
        "dq_init": dq_init_step.DQInitStep,
        "saturation": SaturationStep,
        "refpix": RefPixStep,
        "linearity": LinearityStep,
        "dark_current": DarkCurrentStep,
        "rampfit": ramp_fit_step.RampFitStep,
        "assign_wcs": AssignWcsStep,
        "flatfield": FlatFieldStep,
        "photom": PhotomStep,
        "source_detection": SourceDetectionStep,
        "tweakreg": TweakRegStep,
    }

    # start the actual processing
    def process(self, input):
        """Process the Roman WFI data"""

        log.info("Starting Roman exposure calibration pipeline ...")
        if isinstance(input, str):
            input_filename = basename(input)
        else:
            input_filename = None

        # open the input file
        file_type = filetype.check(input)
        asn = None
        if file_type == "asdf":
            try:
                input = rdm.open(input)
            except TypeError:
                log.debug("Error opening file:")
                return

        if file_type == "asn":
            asn = ModelContainer.read_asn(input)

        # Build a list of observations to process
        expos_file = []
        n_members = 0
        if file_type == "asdf":
            expos_file = [input]
        elif file_type == "asn":
            for product in asn["products"]:
                n_members = len(product["members"])
                for member in product["members"]:
                    expos_file.append(member["expname"])

        results = []
        tweakreg_input = ModelContainer()
        for in_file in expos_file:
            if isinstance(in_file, str):
                input_filename = basename(in_file)
                log.info(f"Input file name: {input_filename}")
            else:
                input_filename = None

            # Open the file
            input = rdm.open(in_file)
            log.info(f"Processing a WFI exposure {in_file}")

            self.dq_init.suffix = "dq_init"
            result = self.dq_init(input)
            if input_filename:
                result.meta.filename = input_filename
            result = self.saturation(result)

            # Test for fully saturated data
            if is_fully_saturated(result):
                log.info("All pixels are saturated. Returning a zeroed-out image.")

                #    if is_fully_saturated(result):
                # Set all subsequent steps to skipped
                for step_str in [
                    "assign_wcs",
                    "flat_field",
                    "photom",
                    "source_detection",
                    "dark",
                    "refpix",
                    "linearity",
                    "ramp_fit",
                    "jump",
                    "tweakreg",
                ]:
                    result.meta.cal_step[step_str] = "SKIPPED"

                # Set suffix for proper output naming
                self.suffix = "cal"
                results.append(result)
                # Return fully saturated image file (stopping pipeline)
                return results

            result = self.refpix(result)
            result = self.linearity(result)
            result = self.dark_current(result)
            result = self.rampfit(result)
            result = self.assign_wcs(result)

            if result.meta.exposure.type == "WFI_IMAGE":
                result = self.flatfield(result)
                result = self.photom(result)
                result = self.source_detection(result)
                if file_type == "asn":
                    result.meta.cal_step.tweakreg = "SKIPPED"
                    self.suffix = "sourcedetection"
                    tweakreg_input.append(result)
                    log.info(
                        f"Number of models to tweakreg:   {len(tweakreg_input._models), n_members}"
                    )
            else:
                log.info("Flat Field step is being SKIPPED")
                log.info("Photom step is being SKIPPED")
                log.info("Source Detection step is being SKIPPED")
                log.info("Tweakreg step is being SKIPPED")
                result.meta.cal_step.flat_field = "SKIPPED"
                result.meta.cal_step.photom = "SKIPPED"
                result.meta.cal_step.source_detection = "SKIPPED"
                result.meta.cal_step.tweakreg = "SKIPPED"
                self.suffix = "cal"

            # setup output_file for saving
            if result.meta.cal_step.tweakreg == "COMPLETE":
                self.suffix = "cal"

            self.setup_output(result)

            self.output_use_model = True
            results.append(result)

        # Now that all the exposures are collated, run tweakreg
        # Note: this does not cover the case where the asn mixes imaging and spectral
        #          observations. This should not occur on-prem
        if result.meta.exposure.type == "WFI_IMAGE":
            if file_type == "asdf":
                result.meta.cal_step.tweakreg = "SKIPPED"
                mc_result = ModelContainer(result)
                if hasattr(ModelContainer(result), "_models") and mc_result._models:
                    result = mc_result._models.pop()
                else:
                    result.meta.cal_step.tweakreg == "SKIPPED"

            if file_type == "asn":
                result = self.tweakreg(tweakreg_input)
                for model in result._models:
                    model.meta.cal_step.tweakreg = "SKIPPED"

            log.info("Roman exposure calibration pipeline ending...")

        return results

    def setup_output(self, input):
        """Determine the proper file name suffix to use later"""
        if input.meta.cal_step.tweakreg == "COMPLETE":
            self.suffix = "cal"
        # elif input.meta.cal_step.tweakreg == "SKIPPED" and input.meta.exposure.type == "WFI_IMAGE":
        #    self.suffix = "sourcedetection"
        elif (
            input.meta.cal_step.tweakreg == "INCOMPLETE"
            and input.meta.exposure.type == "WFI_IMAGE"
        ):
            self.suffix = "sourcedetection"
        elif (
            input.meta.cal_step.tweakreg == "SKIPPED"
            and input.meta.exposure.type != "WFI_IMAGE"
        ):
            self.suffix = "cal"

        if not self.steps["tweakreg"]["skip"]:
            self.suffix = "cal"

        input.meta.filename = input.meta.filename.replace("uncal", self.suffix)
        input["output_file"] = input.meta.filename
        self.output_file = input.meta.filename

    def create_fully_saturated_zeroed_image(self, input_model):
        """
        Create zeroed-out image file
        """
        # The set order is: data, dq, var_poisson, var_rnoise, err
        fully_saturated_model = ramp_fit_step.create_image_model(
            input_model,
            (
                np.zeros(input_model.data.shape[1:], dtype=input_model.data.dtype),
                input_model.pixeldq | input_model.groupdq[0] | group.SATURATED,
                np.zeros(input_model.err.shape[1:], dtype=input_model.err.dtype),
                np.zeros(input_model.err.shape[1:], dtype=input_model.err.dtype),
                np.zeros(input_model.err.shape[1:], dtype=input_model.err.dtype),
            ),
        )

        # Set all subsequent steps to skipped
        for step_str in [
            "linearity",
            "dark",
            "ramp_fit",
            "assign_wcs",
            "flat_field",
            "photom",
            "source_detection",
            "tweakreg",
        ]:
            fully_saturated_model.meta.cal_step[step_str] = "SKIPPED"

        # Set suffix for proper output naming
        self.suffix = "cal"

        # Return zeroed-out image file
        return fully_saturated_model



"""
Module to estimate a 2D background and background RMS noise.
"""

import logging

import numpy as np
from astropy.stats import SigmaClip
from astropy.utils import lazyproperty
from photutils.background import Background2D, MedianBackground

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class RomanBackground:
    """
    Class to estimate a 2D background and background RMS noise in an
    image.

    Parameters
    ----------
    data : 2D `~numpy.ndarray`
        The input 2D array.

    box_size : int or array_like (int)
        The box size along each axis.  If ``box_size`` is a scalar then
        a square box of size ``box_size`` will be used.  If ``box_size``
        has two elements, they should be in ``(ny, nx)`` order.

    mask : array_like (bool), optional
        A boolean mask, with the same shape as ``data``, where a `True`
        value indicates the corresponding element of ``data`` is
        masked. Masked data are excluded from calculations. ``mask``
        and ``coverage_mask`` differ only in that ``coverage_mask`` is
        applied to the output background and background RMS maps.

    coverage_mask : array_like (bool), optional
        A boolean mask, with the same shape as ``data``, where a `True`
        value indicates the corresponding element of ``data`` is masked.
        Masked data are excluded from calculations. ``coverage_mask``
        should be `True` where there is no coverage (i.e., no data) for
        a given pixel (e.g., blank areas in a mosaic image). It should
        not be used for bad pixels.

    Attributes
    ----------
    background : 2D `~numpy.ndimage`
        The estimated 2D background image.

    background_rms : 2D `~numpy.ndimage`
        The estimated 2D background RMS image.
    """

    def __init__(self, data, box_size=100, mask=None, coverage_mask=None):
        self.data = data
        self.box_size = np.asarray(box_size).astype(int)  # must be integer
        self.mask = mask
        self.coverage_mask = coverage_mask

    @lazyproperty
    def _background2d(self):
        """
        Estimate the 2D background and background RMS noise in an image.

        Returns
        -------
        background : `photutils.background.Background2D`
            A Background2D object containing the 2D background and
            background RMS noise estimates.
        """
        sigma_clip = SigmaClip(sigma=3.0)
        bkg_estimator = MedianBackground()
        filter_size = (3, 3)

        try:
            bkg = Background2D(
                self.data,
                self.box_size,
                filter_size=filter_size,
                mask=self.mask,
                coverage_mask=self.coverage_mask,
                sigma_clip=sigma_clip,
                bkg_estimator=bkg_estimator,
            )
        except ValueError:
            # use the entire unmasked array
            bkg = Background2D(
                self.data,
                self.data.shape,
                filter_size=filter_size,
                mask=self.mask,
                coverage_mask=self.coverage_mask,
                sigma_clip=sigma_clip,
                bkg_estimator=bkg_estimator,
                exclude_percentile=100.0,
            )
            log.info(
                "Background could not be estimated in meshes. "
                "Using the entire unmasked array for background "
                f"estimation: bkg_boxsize={self.data.shape}."
            )

        return bkg

    @lazyproperty
    def background(self):
        """
        The 2D background image.
        """
        return self._background2d.background

    @lazyproperty
    def background_rms(self):
        """
        The 2D background RMS image.
        """
        return self._background2d.background_rms



#!/usr/bin/env python

import importlib
import pkgutil

import pytest

import romancal


def dependencies(package, exclude: [str]):
    return [
        module[1]
        for module in pkgutil.walk_packages(
            package.__path__, prefix=package.__name__ + "."
        )
        if not any(exclude_module in module[1] for exclude_module in exclude)
    ]


MODULES = dependencies(romancal, exclude=["test", "time", "static_preview"])


@pytest.mark.filterwarnings("ignore:romancal.dqflags is deprecated.*")
@pytest.mark.parametrize(
    "module_name",
    MODULES,
)
def test_module_import(module_name):
    importlib.import_module(module_name)



import numpy as np
import pytest
from astropy import units as u
from roman_datamodels import maker_utils, stnode
from roman_datamodels.datamodels import MaskRefModel, ScienceRawModel
from roman_datamodels.dqflags import pixel

from romancal.dq_init import DQInitStep
from romancal.dq_init.dq_initialization import do_dqinit

RNG = np.random.default_rng(83)

# Set parameters for multiple runs of data
args = "xstart, ystart, xsize, ysize, ngroups, instrument, exp_type"
test_data = [(1, 1, 2048, 2048, 2, "WFI", "WFI_IMAGE")]
ids = ["RampModel"]


@pytest.mark.parametrize(args, test_data, ids=ids)
def test_dq_im(xstart, ystart, xsize, ysize, ngroups, instrument, exp_type):
    """Check that PIXELDQ is initialized with the information from the reference file.
    test that a flagged value in the reference file flags the PIXELDQ array"""

    csize = (ngroups, ysize, xsize)

    # create raw input data for step
    dm_ramp = maker_utils.mk_ramp(shape=csize)
    dm_ramp.meta.instrument.name = instrument

    # create a MaskModel elements for the dq input mask
    dq = np.zeros(csize[1:], dtype=np.uint32)

    # edit reference file with known bad pixel values
    dq[100, 100] = 2  # Saturated pixel
    dq[200, 100] = 4  # Jump detected pixel
    dq[300, 100] = 8  # Dropout
    dq[400, 100] = 32  # Persistence
    dq[500, 100] = 1  # Do_not_use
    dq[600, 100] = 16  # guide window affected data
    dq[100, 200] = 3  # Saturated pixel + do not use
    dq[200, 200] = 5  # Jump detected pixel + do not use
    dq[300, 200] = 9  # Dropout + do not use
    dq[400, 200] = 33  # Persistence + do not use

    # write mask model
    ref_data = maker_utils.mk_mask(shape=csize[1:])

    # Copy in maskmodel elemnts
    ref_data["dq"] = dq

    ref_data["meta"]["instrument"]["name"] = instrument

    # run do_dqinit
    outfile = do_dqinit(dm_ramp, ref_data)
    dqdata = outfile["pixeldq"]

    # assert that the pixels read back in match the mapping from ref data to
    # science data
    assert dqdata[100, 100] == pixel.SATURATED
    assert dqdata[200, 100] == pixel.JUMP_DET
    assert dqdata[300, 100] == pixel.DROPOUT
    assert dqdata[400, 100] == pixel.PERSISTENCE
    assert dqdata[500, 100] == pixel.DO_NOT_USE
    assert dqdata[600, 100] == pixel.GW_AFFECTED_DATA
    assert dqdata[100, 200] == pixel.SATURATED + pixel.DO_NOT_USE
    assert dqdata[200, 200] == pixel.JUMP_DET + pixel.DO_NOT_USE
    assert dqdata[300, 200] == pixel.DROPOUT + pixel.DO_NOT_USE
    assert dqdata[400, 200] == pixel.PERSISTENCE + pixel.DO_NOT_USE


def test_groupdq():
    """
    Check that GROUPDQ extension is added to the data and all values are
    initialized to zero.
    """

    # size of integration
    instrument = "WFI"
    ngroups = 5
    xsize = 1032
    ysize = 1024
    csize = (ngroups, ysize, xsize)

    # create raw input data for step
    dm_ramp = maker_utils.mk_ramp(shape=csize)
    dm_ramp.meta.instrument.name = instrument

    # create a MaskModel elements for the dq input mask
    ref_data = maker_utils.mk_mask(shape=csize[1:])
    ref_data["meta"]["instrument"]["name"] = instrument

    # run the correction step
    outfile = do_dqinit(dm_ramp, ref_data)

    # check that GROUPDQ was created and initialized to zero
    groupdq = outfile.groupdq

    np.testing.assert_array_equal(
        np.full((ngroups, ysize, xsize), 0, dtype=int),
        groupdq,
        err_msg="groupdq not initialized to zero",
    )


def test_err():
    """Check that a 3-D ERR array is initialized and all values are zero."""

    # size of integration
    instrument = "WFI"
    ngroups = 5
    xsize = 1032
    ysize = 1024
    csize = (ngroups, ysize, xsize)

    # create raw input data for step
    dm_ramp = maker_utils.mk_ramp(shape=(ngroups, ysize, xsize))
    dm_ramp.meta.instrument.name = instrument

    # create a MaskModel elements for the dq input mask
    ref_data = maker_utils.mk_mask(shape=csize[1:])
    ref_data["meta"]["instrument"]["name"] = instrument

    # run correction step
    outfile = do_dqinit(dm_ramp, ref_data)

    # check that ERR array was created and initialized to zero
    errarr = outfile.err

    assert errarr.ndim == 3  # check that output err array is 3-D
    assert np.all(errarr == 0)  # check that values are 0


def test_dq_add1_groupdq():
    """
    Test if the dq_init code set the groupdq flag on the first
    group to 'do_not_use' by adding 1 to the flag, not overwriting to 1
    Also test whether two flags on the same pixel are added together.
    """

    # size of integration
    instrument = "WFI"
    ngroups = 5
    xsize = 1032
    ysize = 1024
    csize = (ngroups, ysize, xsize)

    # create raw input data for step
    dm_ramp = maker_utils.mk_ramp(shape=(ngroups, ysize, xsize))
    dm_ramp.meta.instrument.name = instrument

    # create a MaskModel elements for the dq input mask
    dq = np.zeros(csize[1:], dtype=np.uint32)

    # write reference file with known bad pixel values
    dq[505, 505] = 1  # Do_not_use
    dq[400, 500] = 3  # do_not_use and saturated pixel

    # write mask model
    ref_data = maker_utils.mk_mask(shape=csize[1:])
    ref_data["meta"]["instrument"]["name"] = instrument

    # Copy in maskmodel elemnts
    ref_data["dq"] = dq

    # set a flag in the pixel dq
    dm_ramp.pixeldq[505, 505] = 4  # Jump detected pixel

    # run correction step
    outfile = do_dqinit(dm_ramp, ref_data)

    # test if pixels in pixeldq were incremented in value by 1
    # check that previous dq flag is added to mask value
    assert outfile.pixeldq[505, 505] == pixel.JUMP_DET + pixel.DO_NOT_USE
    # check two flags propagate correctly
    assert outfile.pixeldq[400, 500] == pixel.SATURATED + pixel.DO_NOT_USE


@pytest.mark.parametrize(
    "instrument, exptype",
    [
        ("WFI", "WFI_IMAGE"),
    ],
)
def test_dqinit_step_interface(instrument, exptype):
    """Test that the basic inferface works for data requiring a DQ reffile"""

    # Set test size
    shape = (2, 20, 20)

    # include an extra item not defined in the schema
    extra_key = "foo_extra"
    extra_value = [1, 2, 3]

    # Create test science raw model
    wfi_sci_raw = maker_utils.mk_level1_science_raw(shape=shape)
    wfi_sci_raw.meta.instrument.name = instrument
    wfi_sci_raw.meta.instrument.detector = "WFI01"
    wfi_sci_raw.meta.instrument.optical_element = "F158"
    wfi_sci_raw.meta["guidestar"]["gw_window_xstart"] = 1012
    wfi_sci_raw.meta["guidestar"]["gw_window_xsize"] = 16
    wfi_sci_raw.meta.exposure.type = exptype
    wfi_sci_raw.data = u.Quantity(
        np.ones(shape, dtype=np.uint16), u.DN, dtype=np.uint16
    )
    wfi_sci_raw[extra_key] = extra_value
    wfi_sci_raw_model = ScienceRawModel(wfi_sci_raw)

    # Create mask model
    maskref = stnode.MaskRef()
    meta = maker_utils.mk_ref_common("MASK")
    meta["instrument"]["optical_element"] = "F158"
    meta["instrument"]["detector"] = "WFI01"
    maskref["meta"] = meta
    maskref["data"] = np.ones(shape[1:], dtype=np.float32)
    maskref["dq"] = np.zeros(shape[1:], dtype=np.uint32)
    maskref["err"] = (RNG.uniform(size=shape[1:]) * 0.05).astype(np.float32)
    maskref_model = MaskRefModel(maskref)

    # Perform Data Quality application step
    result = DQInitStep.call(wfi_sci_raw_model, override_mask=maskref_model)

    # Test dq_init results
    assert (result.data == wfi_sci_raw.data).all()
    assert result.pixeldq.shape == shape[1:]
    assert result.meta.cal_step.dq_init == "COMPLETE"
    assert result.data.dtype == np.float32
    assert result.err.dtype == np.float32
    assert result.pixeldq.dtype == np.uint32
    assert result.groupdq.dtype == np.uint8
    # check that extra value came through
    assert result[extra_key] == extra_value


@pytest.mark.parametrize(
    "instrument, exptype",
    [
        ("WFI", "WFI_IMAGE"),
    ],
)
def test_dqinit_refpix(instrument, exptype):
    """Test that the basic inferface works for data requiring a DQ reffile"""

    # Set test size
    shape = (2, 20, 20)

    # Create test science raw model
    wfi_sci_raw = maker_utils.mk_level1_science_raw(shape=shape)
    wfi_sci_raw.meta.instrument.name = instrument
    wfi_sci_raw.meta.instrument.detector = "WFI01"
    wfi_sci_raw.meta.instrument.optical_element = "F158"
    wfi_sci_raw.meta["guidestar"]["gw_window_xstart"] = 1012
    wfi_sci_raw.meta["guidestar"]["gw_window_xsize"] = 16
    wfi_sci_raw.meta.exposure.type = exptype
    wfi_sci_raw.data = u.Quantity(
        np.ones(shape, dtype=np.uint16), u.DN, dtype=np.uint16
    )
    wfi_sci_raw_model = ScienceRawModel(wfi_sci_raw)

    # Create mask model
    maskref = stnode.MaskRef()
    meta = maker_utils.mk_ref_common("MASK")
    meta["instrument"]["optical_element"] = "F158"
    meta["instrument"]["detector"] = "WFI01"
    maskref["meta"] = meta
    maskref["data"] = np.ones(shape[1:], dtype=np.float32)
    maskref["dq"] = np.zeros(shape[1:], dtype=np.uint32)
    maskref["err"] = (RNG.uniform(size=shape[1:]) * 0.05).astype(np.float32)
    maskref_model = MaskRefModel(maskref)

    # Perform Data Quality application step
    result = DQInitStep.call(wfi_sci_raw_model, override_mask=maskref_model)

    # check if reference pixels are correct
    assert result.data.shape == (2, 20, 20)  # no pixels should be trimmed
    assert result.amp33.value.shape == (2, 4096, 128)
    assert result.border_ref_pix_right.shape == (2, 20, 4)
    assert result.border_ref_pix_left.shape == (2, 20, 4)
    assert result.border_ref_pix_top.shape == (2, 4, 20)
    assert result.border_ref_pix_bottom.shape == (2, 4, 20)
    assert result.dq_border_ref_pix_right.shape == (20, 4)
    assert result.dq_border_ref_pix_left.shape == (20, 4)
    assert result.dq_border_ref_pix_top.shape == (4, 20)
    assert result.dq_border_ref_pix_bottom.shape == (4, 20)


@pytest.mark.parametrize(
    "instrument, exptype",
    [
        ("WFI", "WFI_IMAGE"),
    ],
)
def test_dqinit_resultantdq(instrument, exptype):
    """Test that the basic inferface works for data requiring a DQ reffile"""

    # Set test size
    shape = (2, 20, 20)

    # Create test science raw model
    wfi_sci_raw = maker_utils.mk_level1_science_raw(shape=shape, dq=True)
    wfi_sci_raw.meta.instrument.name = instrument
    wfi_sci_raw.meta.instrument.detector = "WFI01"
    wfi_sci_raw.meta.instrument.optical_element = "F158"
    wfi_sci_raw.meta["guidestar"]["gw_window_xstart"] = 1012
    wfi_sci_raw.meta["guidestar"]["gw_window_xsize"] = 16
    wfi_sci_raw.meta.exposure.type = exptype
    wfi_sci_raw.resultantdq[1, 12, 12] = pixel["DROPOUT"]
    wfi_sci_raw.data = u.Quantity(
        np.ones(shape, dtype=np.uint16), u.DN, dtype=np.uint16
    )
    wfi_sci_raw_model = ScienceRawModel(wfi_sci_raw)

    # Create mask model
    maskref = stnode.MaskRef()
    meta = maker_utils.mk_ref_common("MASK")
    meta["instrument"]["optical_element"] = "F158"
    meta["instrument"]["detector"] = "WFI01"
    maskref["meta"] = meta
    maskref["data"] = np.ones(shape[1:], dtype=np.float32)
    maskref["dq"] = np.zeros(shape[1:], dtype=np.uint32)
    maskref["err"] = (RNG.uniform(size=shape[1:]) * 0.05).astype(np.float32)
    maskref_model = MaskRefModel(maskref)

    # Perform Data Quality application step
    result = DQInitStep.call(wfi_sci_raw_model, override_mask=maskref_model)

    # check that the resultantdq is present in the raw model
    assert hasattr(wfi_sci_raw_model, "resultantdq")
    # check that the resultantdq is not copied to the result
    assert not hasattr(result, "resultantdq")
    # check to see the resultantdq is the correct shape
    assert wfi_sci_raw_model.resultantdq.shape == shape
    # check to see the resultantdq & groupdq have the correct value
    assert wfi_sci_raw_model.resultantdq[1, 12, 12] == pixel["DROPOUT"]
    assert result.groupdq[1, 12, 12] == pixel["DROPOUT"]


@pytest.mark.parametrize(
    "instrument, exptype",
    [
        ("WFI", "WFI_IMAGE"),
    ],
)
def test_dqinit_getbestref(instrument, exptype):
    """Test that the step is skipped if the CRDS reffile is N/A"""

    # Set test size
    shape = (2, 20, 20)

    # Create test science raw model
    wfi_sci_raw = maker_utils.mk_level1_science_raw(shape=shape)
    wfi_sci_raw.meta.instrument.name = instrument
    wfi_sci_raw.meta.instrument.detector = "WFI01"
    wfi_sci_raw.meta.instrument.optical_element = "F158"
    wfi_sci_raw.meta["guidestar"]["gw_window_xstart"] = 1012
    wfi_sci_raw.meta["guidestar"]["gw_window_xsize"] = 16
    wfi_sci_raw.meta.exposure.type = exptype
    wfi_sci_raw.data = u.Quantity(
        np.ones(shape, dtype=np.uint16), u.DN, dtype=np.uint16
    )
    wfi_sci_raw_model = ScienceRawModel(wfi_sci_raw)

    # Perform Data Quality application step
    result = DQInitStep.call(wfi_sci_raw_model, override_mask="N/A")
    assert result.meta.cal_step.dq_init == "SKIPPED"



"""
The ``skyimage`` module contains algorithms that are used by
``skymatch`` to manage all of the information for footprints (image outlines)
on the sky as well as perform useful operations on these outlines such as
computing intersections and statistics in the overlap regions.
"""

import abc
import pickle  # nosec B403
import tempfile

import numpy as np
from spherical_geometry.polygon import SphericalPolygon

from . import region
from .skystatistics import SkyStats

__all__ = [
    "SkyImage",
    "SkyGroup",
    "DataAccessor",
    "NDArrayInMemoryAccessor",
    "NDArrayMappedAccessor",
]


class DataAccessor(abc.ABC):
    """Base class for all data accessors. Provides a common interface to
    access data.
    """

    @abc.abstractmethod
    def get_data(self):
        pass

    @abc.abstractmethod
    def set_data(self, data):
        """Sets data.

        Parameters
        ----------
        data : numpy.ndarray
            Data array to be set.

        """
        pass

    @abc.abstractmethod
    def get_data_shape(self):
        pass


class NDArrayInMemoryAccessor(DataAccessor):
    """Acessor for in-memory `numpy.ndarray` data."""

    def __init__(self, data):
        super().__init__()
        self._data = data

    def get_data(self):
        return self._data

    def set_data(self, data):
        self._data = data

    def get_data_shape(self):
        return np.shape(self._data)


class NDArrayMappedAccessor(DataAccessor):
    """Data accessor for arrays stored in temporary files."""

    def __init__(
        self, data, tmpfile=None, prefix="tmp_skymatch_", suffix=".npy", tmpdir=""
    ):
        super().__init__()
        if tmpfile is None:
            self._close = True
            self._tmp = tempfile.NamedTemporaryFile(
                prefix=prefix, suffix=suffix, dir=tmpdir
            )
            if not self._tmp:
                raise RuntimeError("Unable to create temporary file.")
        else:
            # temp file managed by the caller
            self._close = False
            self._tmp = tmpfile

        self.set_data(data)

    def get_data(self):
        self._tmp.seek(0)
        return pickle.load(self._tmp)  # nosec B301

    def set_data(self, data):
        data = np.asanyarray(data)
        self._data_shape = data.shape
        self._tmp.seek(0)
        pickle.dump(data, self._tmp)

    def __del__(self):
        if self._close:
            self._tmp.close()

    def get_data_shape(self):
        return self._data_shape


class SkyImage:
    """
    Container that holds information about properties of a *single*
    image such as:

    * image data;
    * WCS of the chip image;
    * bounding spherical polygon;
    * id;
    * pixel area;
    * sky background value;
    * sky statistics parameters;
    * mask associated image data indicating "good" (1) data.

    """

    def __init__(
        self,
        image,
        wcs_fwd,
        wcs_inv,
        pix_area=1.0,
        convf=1.0,
        mask=None,
        id=None,
        skystat=None,
        stepsize=None,
        meta=None,
        reduce_memory_usage=True,
    ):
        """Initializes the SkyImage object.

        Parameters
        ----------
        image : numpy.ndarray, NDArrayDataAccessor
            A 2D array of image data or a `NDArrayDataAccessor`.

        wcs_fwd : function
            "forward" pixel-to-world transformation function.

        wcs_inv : function
            "inverse" world-to-pixel transformation function.

        pix_area : float, optional
            Average pixel's sky area.

        convf : float, optional
            Conversion factor that when multiplied to `image` data converts
            the data to "uniform" (across multiple images) surface
            brightness units.

            .. note::

              The functionality to support this conversion is not yet
              implemented and at this moment `convf` is ignored.

        mask : numpy.ndarray, NDArrayDataAccessor
            A 2D array or `NDArrayDataAccessor` of a 2D array that indicates
            which pixels in the input `image` should be used for sky
            computations (``1``) and which pixels should **not** be used
            for sky computations (``0``).

        id : anything
            The value of this parameter is simply stored within the `SkyImage`
            object. While it can be of any type, it is preferable that `id` be
            of a type with nice string representation.

        skystat : callable, None, optional
            A callable object that takes a either a 2D image (2D
            `numpy.ndarray`) or a list of pixel values (an Nx1 array) and
            returns a tuple of two values: some statistics (e.g., mean,
            median, etc.) and number of pixels/values from the input image
            used in computing that statistics.

            When `skystat` is not set, `SkyImage` will use
            :py:class:`~romancal.skymatch.skystatistics.SkyStats` object
            to perform sky statistics on image data.

        stepsize : int, None, optional
            Spacing between vertices of the image's bounding polygon. Default
            value of `None` creates bounding polygons with four vertices
            corresponding to the corners of the image.

        meta : dict, None, optional
            A dictionary of various items to be stored within the `SkyImage`
            object.

        reduce_memory_usage : bool, optional
            Indicates whether to attempt to minimize memory usage by attaching
            input ``image`` and/or ``mask`` `numpy.ndarray` arrays to
            file-mapped accessor. This has no effect when input parameters
            ``image`` and/or ``mask`` are already of `NDArrayDataAccessor`
            objects.

        """
        self._image = None
        self._mask = None
        self._image_shape = None
        self._mask_shape = None
        self._reduce_memory_usage = reduce_memory_usage

        self.image = image

        self.convf = convf
        self.meta = meta
        self._id = id
        self._pix_area = pix_area

        # WCS
        self.wcs_fwd = wcs_fwd
        self.wcs_inv = wcs_inv

        # initial sky value:
        self._sky = 0.0
        self._sky_is_valid = False

        self.mask = mask

        # create spherical polygon bounding the image
        if image is None or wcs_fwd is None or wcs_inv is None:
            self._radec = [(np.array([]), np.array([]))]
            self._polygon = SphericalPolygon([])
            self._poly_area = 0.0

        else:
            self.calc_bounding_polygon(stepsize)

        # set sky statistics function (NOTE: it must return statistics and
        # the number of pixels used after clipping)
        if skystat is None:
            self.set_builtin_skystat()
        else:
            self.skystat = skystat

    @property
    def mask(self):
        """Set or get `SkyImage`'s ``mask`` data array or `None`."""
        if self._mask is None:
            return None
        else:
            return self._mask.get_data()

    @mask.setter
    def mask(self, mask):
        if mask is None:
            self._mask = None
            self._mask_shape = None

        elif isinstance(mask, DataAccessor):
            if self._image is None:
                raise ValueError("'mask' must be None when 'image' is None")

            self._mask = mask
            self._mask_shape = mask.get_data_shape()

            # check that mask has the same shape as image:
            if self._mask_shape != self.image_shape:
                raise ValueError("'mask' must have the same shape as 'image'.")

        else:
            if self._image is None:
                raise ValueError("'mask' must be None when 'image' is None")

            mask = np.asanyarray(mask, dtype=bool)
            self._mask_shape = mask.shape

            # check that mask has the same shape as image:
            if self._mask_shape != self.image_shape:
                raise ValueError("'mask' must have the same shape as 'image'.")

            if self._mask is None:
                if self._reduce_memory_usage:
                    self._mask = NDArrayMappedAccessor(
                        mask, prefix="tmp_skymatch_mask_"
                    )
                else:
                    self._mask = NDArrayInMemoryAccessor(mask)
            else:
                self._mask.set_data(mask)

    @property
    def image(self):
        """Set or get `SkyImage`'s ``image`` data array."""
        if self._image is None:
            return None
        else:
            return self._image.get_data()

    @image.setter
    def image(self, image):
        if image is None:
            self._image = None
            self._image_shape = None
            self.mask = None

        if isinstance(image, DataAccessor):
            self._image = image
            self._image_shape = image.get_data_shape()

        else:
            image = np.asanyarray(image)
            self._image_shape = image.shape
            if self._image is None:
                if self._reduce_memory_usage:
                    self._image = NDArrayMappedAccessor(
                        image, prefix="tmp_skymatch_image_"
                    )
                else:
                    self._image = NDArrayInMemoryAccessor(image)
            else:
                self._image.set_data(image)

    @property
    def image_shape(self):
        """Get `SkyImage`'s ``image`` data shape."""
        if self._image_shape is None and self._image is not None:
            self._image_shape = self._image.get_data_shape()
        return self._image_shape

    @property
    def id(self):
        """Set or get `SkyImage`'s `id`.

        While `id` can be of any type, it is preferable that `id` be
        of a type with nice string representation.

        """
        return self._id

    @id.setter
    def id(self, id):
        self._id = id

    @property
    def pix_area(self):
        """Set or get mean pixel area."""
        return self._pix_area

    @pix_area.setter
    def pix_area(self, pix_area):
        self._pix_area = pix_area

    @property
    def poly_area(self):
        """Get bounding polygon area in srad units."""
        return self._poly_area

    @property
    def sky(self):
        """Sky background value. See `calc_sky` for more details."""
        return self._sky

    @sky.setter
    def sky(self, sky):
        self._sky = sky

    @property
    def is_sky_valid(self):
        """
        Indicates whether sky value was successfully computed.
        Must be set externally.
        """
        return self._sky_is_valid

    @is_sky_valid.setter
    def is_sky_valid(self, valid):
        self._sky_is_valid = valid

    @property
    def radec(self):
        """
        Get RA and DEC of the vertices of the bounding polygon as a
        `~numpy.ndarray` of shape (N, 2) where N is the number of vertices + 1.
        """
        return self._radec

    @property
    def polygon(self):
        """Get image's bounding polygon."""
        return self._polygon

    def intersection(self, skyimage):
        """
        Compute intersection of this `SkyImage` object and another
        `SkyImage`, `SkyGroup`, or
        :py:class:`~spherical_geometry.polygon.SphericalPolygon`
        object.

        Parameters
        ----------
        skyimage : SkyImage, SkyGroup, SphericalPolygon
            Another object that should be intersected with this `SkyImage`.

        Returns
        -------
        polygon : SphericalPolygon
            A :py:class:`~spherical_geometry.polygon.SphericalPolygon` that is
            the intersection of this `SkyImage` and `skyimage`.

        """
        if isinstance(skyimage, (SkyImage, SkyGroup)):
            other = skyimage.polygon
        else:
            other = skyimage

        pts1 = np.sort(list(self._polygon.points)[0], axis=0)
        pts2 = np.sort(list(other.points)[0], axis=0)
        if np.allclose(pts1, pts2, rtol=0, atol=5e-9):
            intersect_poly = self._polygon.copy()
        else:
            intersect_poly = self._polygon.intersection(other)
        return intersect_poly

    def calc_bounding_polygon(self, stepsize=None):
        """Compute image's bounding polygon.

        Parameters
        ----------
        stepsize : int, None, optional
            Indicates the maximum separation between two adjacent vertices
            of the bounding polygon along each side of the image. Corners
            of the image are included automatically. If `stepsize` is `None`,
            bounding polygon will contain only vertices of the image.

        """
        ny, nx = self.image_shape

        if stepsize is None:
            nintx = 2
            ninty = 2
        else:
            nintx = max(2, int(np.ceil((nx + 1.0) / stepsize)))
            ninty = max(2, int(np.ceil((ny + 1.0) / stepsize)))

        xs = np.linspace(-0.5, nx - 0.5, nintx, dtype=float)
        ys = np.linspace(-0.5, ny - 0.5, ninty, dtype=float)[1:-1]
        nptx = xs.size
        npty = ys.size

        npts = 2 * (nptx + npty)

        borderx = np.empty((npts + 1,), dtype=float)
        bordery = np.empty((npts + 1,), dtype=float)

        # "bottom" points:
        borderx[:nptx] = xs
        bordery[:nptx] = -0.5
        # "right"
        sl = np.s_[nptx : nptx + npty]
        borderx[sl] = nx - 0.5
        bordery[sl] = ys
        # "top"
        sl = np.s_[nptx + npty : 2 * nptx + npty]
        borderx[sl] = xs[::-1]
        bordery[sl] = ny - 0.5
        # "left"
        sl = np.s_[2 * nptx + npty : -1]
        borderx[sl] = -0.5
        bordery[sl] = ys[::-1]

        # close polygon:
        borderx[-1] = borderx[0]
        bordery[-1] = bordery[0]

        ra, dec = self.wcs_fwd(borderx, bordery, with_bounding_box=False)
        # TODO: for strange reasons, occasionally ra[0] != ra[-1] and/or
        #       dec[0] != dec[-1] (even though we close the polygon in the
        #       previous two lines). Then SphericalPolygon fails because
        #       points are not closed. Therefore we force it to be closed:
        ra[-1] = ra[0]
        dec[-1] = dec[0]

        self._radec = [(ra, dec)]
        self._polygon = SphericalPolygon.from_radec(ra, dec)
        self._poly_area = np.fabs(self._polygon.area())

    @property
    def skystat(self):
        """Stores/retrieves a callable object that takes a either a 2D image
        (2D `numpy.ndarray`) or a list of pixel values (an Nx1 array) and
        returns a tuple of two values: some statistics
        (e.g., mean, median, etc.) and number of pixels/values from the input
        image used in computing that statistics.

        When `skystat` is not set, `SkyImage` will use
        :py:class:`~romancal.skymatch.skystatistics.SkyStats` object
        to perform sky statistics on image data.

        """
        return self._skystat

    @skystat.setter
    def skystat(self, skystat):
        self._skystat = skystat

    def set_builtin_skystat(
        self,
        skystat="median",
        lower=None,
        upper=None,
        nclip=5,
        lsigma=4.0,
        usigma=4.0,
        binwidth=0.1,
    ):
        """
        Replace already set `skystat` with a "built-in" version of a
        statistics callable object used to measure sky background.

        See :py:class:`~romancal.skymatch.skystatistics.SkyStats` for the
        parameter description.

        """
        self._skystat = SkyStats(
            skystat=skystat,
            lower=lower,
            upper=upper,
            nclip=nclip,
            lsig=lsigma,
            usig=usigma,
            binwidth=binwidth,
        )

    def calc_sky(self, overlap=None, delta=True):
        """
        Compute sky background value.

        Parameters
        ----------
        overlap : SkyImage, SkyGroup, SphericalPolygon, list of tuples, \
None, optional
            Another `SkyImage`, `SkyGroup`,
            :py:class:`spherical_geometry.polygons.SphericalPolygon`, or
            a list of tuples of (RA, DEC) of vertices of a spherical
            polygon. This parameter is used to indicate that sky statistics
            should computed only in the region of intersection of *this*
            image with the polygon indicated by `overlap`. When `overlap` is
            `None`, sky statistics will be computed over the entire image.

        delta : bool, optional
            Should this function return absolute sky value or the difference
            between the computed value and the value of the sky stored in the
            `sky` property.

        Returns
        -------
        skyval : float, None
            Computed sky value (absolute or relative to the `sky` attribute).
            If there are no valid data to perform this computations (e.g.,
            because this image does not overlap with the image indicated by
            `overlap`), `skyval` will be set to `None`.

        npix : int
            Number of pixels used to compute sky statistics.

        polyarea : float
            Area (in srad) of the polygon that bounds data used to compute
            sky statistics.

        """
        if overlap is None:
            if self._mask is None:
                data = self.image
            else:
                data = self.image[self._mask.get_data()]

            polyarea = self.poly_area

        else:
            fill_mask = np.zeros(self.image_shape, dtype=bool)

            if isinstance(overlap, SkyImage):
                intersection = self.intersection(overlap)
                polyarea = np.fabs(intersection.area())
                radec = list(intersection.to_radec())

            elif isinstance(overlap, SkyGroup):
                radec = []
                polyarea = 0.0
                for im in overlap:
                    intersection = self.intersection(im)
                    polyarea1 = np.fabs(intersection.area())
                    if polyarea1 == 0.0:
                        continue
                    polyarea += polyarea1
                    radec += list(intersection.to_radec())

            elif isinstance(overlap, SphericalPolygon):
                radec = []
                polyarea = 0.0
                for p in overlap._polygons:
                    intersection = self.intersection(SphericalPolygon([p]))
                    polyarea1 = np.fabs(intersection.area())
                    if polyarea1 == 0.0:
                        continue
                    polyarea += polyarea1
                    radec += list(intersection.to_radec())

            else:  # assume a list of (ra, dec) tuples:
                radec = []
                polyarea = 0.0
                for r, d in overlap:
                    poly = SphericalPolygon.from_radec(r, d)
                    polyarea1 = np.fabs(poly.area())
                    if polyarea1 == 0.0 or len(r) < 4:
                        continue
                    polyarea += polyarea1
                    radec.append(self.intersection(poly).to_radec())

            if polyarea == 0.0:
                return None, 0, 0.0

            for ra, dec in radec:
                if len(ra) < 4:
                    continue

                # set pixels in 'fill_mask' that are inside a polygon to True:
                x, y = self.wcs_inv(ra, dec)
                poly_vert = list(zip(*[x, y]))

                polygon = region.Polygon(True, poly_vert)
                fill_mask = polygon.scan(fill_mask)

            if self._mask is not None:
                fill_mask &= self._mask.get_data()

            data = self.image[fill_mask]

            if data.size < 1:
                return None, 0, 0.0

        # Calculate sky
        try:
            skyval, npix = self._skystat(data)
        except ValueError:
            return None, 0, 0.0

        if not np.isfinite(skyval):
            return None, 0, 0.0

        if delta:
            skyval -= self._sky

        return skyval, npix, polyarea

    def _calc_sky_orig(self, overlap=None, delta=True):
        """
        Compute sky background value.

        Parameters
        ----------
        overlap : SkyImage, SkyGroup, SphericalPolygon, list of tuples, \
None, optional
            Another `SkyImage`, `SkyGroup`,
            :py:class:`spherical_geometry.polygons.SphericalPolygon`, or
            a list of tuples of (RA, DEC) of vertices of a spherical
            polygon. This parameter is used to indicate that sky statistics
            should computed only in the region of intersection of *this*
            image with the polygon indicated by `overlap`. When `overlap` is
            `None`, sky statistics will be computed over the entire image.

        delta : bool, optional
            Should this function return absolute sky value or the difference
            between the computed value and the value of the sky stored in the
            `sky` property.

        Returns
        -------
        skyval : float, None
            Computed sky value (absolute or relative to the `sky` attribute).
            If there are no valid data to perform this computations (e.g.,
            because this image does not overlap with the image indicated by
            `overlap`), `skyval` will be set to `None`.

        npix : int
            Number of pixels used to compute sky statistics.

        polyarea : float
            Area (in srad) of the polygon that bounds data used to compute
            sky statistics.

        """

        if overlap is None:
            if self._mask is None:
                data = self.image
            else:
                data = self.image[self._mask.get_data()]

            polyarea = self.poly_area

        else:
            fill_mask = np.zeros(self.image_shape, dtype=bool)

            if isinstance(overlap, (SkyImage, SkyGroup, SphericalPolygon)):
                intersection = self.intersection(overlap)
                polyarea = np.fabs(intersection.area())
                radec = intersection.to_radec()

            else:  # assume a list of (ra, dec) tuples:
                radec = []
                polyarea = 0.0
                for r, d in overlap:
                    poly = SphericalPolygon.from_radec(r, d)
                    polyarea1 = np.fabs(poly.area())
                    if polyarea1 == 0.0 or len(r) < 4:
                        continue
                    polyarea += polyarea1
                    radec.append(self.intersection(poly).to_radec())

            if polyarea == 0.0:
                return None, 0, 0.0

            for ra, dec in radec:
                if len(ra) < 4:
                    continue

                # set pixels in 'fill_mask' that are inside a polygon to True:
                x, y = self.wcs_inv(ra, dec)
                poly_vert = list(zip(*[x, y]))

                polygon = region.Polygon(True, poly_vert)
                fill_mask = polygon.scan(fill_mask)

            if self._mask is not None:
                fill_mask &= self._mask.get_data()

            data = self.image[fill_mask]

            if data.size < 1:
                return None, 0, 0.0

        # Calculate sky
        try:
            skyval, npix = self._skystat(data)

        except ValueError:
            return None, 0, 0.0

        if delta:
            skyval -= self._sky

        return skyval, npix, polyarea

    def copy(self):
        """
        Return a shallow copy of the `SkyImage` object.
        """
        si = SkyImage(
            image=None,
            wcs_fwd=self.wcs_fwd,
            wcs_inv=self.wcs_inv,
            pix_area=self.pix_area,
            convf=self.convf,
            mask=None,
            id=self.id,
            stepsize=None,
            meta=self.meta,
        )

        si._image = self._image
        si._mask = self._mask
        si._image_shape = self._image_shape
        si._mask_shape = self._mask_shape
        si._reduce_memory_usage = self._reduce_memory_usage

        si._radec = self._radec
        si._polygon = self._polygon
        si._poly_area = self._poly_area
        si.sky = self.sky
        return si


class SkyGroup:
    """
    Holds multiple :py:class:`SkyImage` objects whose sky background values
    must be adjusted together.

    `SkyGroup` provides methods for obtaining bounding polygon of the group
    of :py:class:`SkyImage` objects and to compute sky value of the group.

    """

    def __init__(self, images, id=None, sky=0.0):
        if isinstance(images, SkyImage):
            self._images = [images]

        elif hasattr(images, "__iter__"):
            self._images = []
            for im in images:
                if not isinstance(im, SkyImage):
                    raise TypeError(
                        "Each element of the 'images' parameter "
                        "must be an 'SkyImage' object."
                    )
                self._images.append(im)

        else:
            raise TypeError(
                "Parameter 'images' must be either a single "
                "'SkyImage' object or a list of 'SkyImage' objects"
            )

        self._id = id
        self._update_bounding_polygon()
        self._sky = sky
        for im in self._images:
            im.sky += sky

    @property
    def id(self):
        """Set or get `SkyImage`'s `id`.

        While `id` can be of any type, it is preferable that `id` be
        of a type with nice string representation.

        """
        return self._id

    @id.setter
    def id(self, id):
        self._id = id

    @property
    def sky(self):
        """Sky background value. See `calc_sky` for more details."""
        return self._sky

    @sky.setter
    def sky(self, sky):
        delta_sky = sky - self._sky
        self._sky = sky
        for im in self._images:
            im.sky += delta_sky

    @property
    def radec(self):
        """
        Get RA and DEC of the vertices of the bounding polygon as a
        `~numpy.ndarray` of shape (N, 2) where N is the number of vertices + 1.

        """
        return self._radec

    @property
    def polygon(self):
        """Get image's bounding polygon."""
        return self._polygon

    def intersection(self, skyimage):
        """
        Compute intersection of this `SkyImage` object and another
        `SkyImage`, `SkyGroup`, or
        :py:class:`~spherical_geometry.polygon.SphericalPolygon`
        object.

        Parameters
        ----------
        skyimage : SkyImage, SkyGroup, SphericalPolygon
            Another object that should be intersected with this `SkyImage`.

        Returns
        -------
        intersect_poly : SphericalPolygon
            A :py:class:`~spherical_geometry.polygon.SphericalPolygon` that is
            the intersection of this `SkyImage` and `skyimage`.

        """
        if isinstance(skyimage, (SkyImage, SkyGroup)):
            other = skyimage.polygon
        else:
            other = skyimage

        pts1 = np.sort(list(self._polygon.points)[0], axis=0)
        pts2 = np.sort(list(other.points)[0], axis=0)
        if np.allclose(pts1, pts2, rtol=0, atol=1e-8):
            intersect_poly = self._polygon.copy()
        else:
            intersect_poly = self._polygon.intersection(other)
        return intersect_poly

    def _update_bounding_polygon(self):
        polygons = [im.polygon for im in self._images]
        if len(polygons) == 0:
            self._polygon = SphericalPolygon([])
            self._radec = []
        else:
            self._polygon = SphericalPolygon.multi_union(polygons)
            self._radec = list(self._polygon.to_radec())

    def __len__(self):
        return len(self._images)

    def __getitem__(self, idx):
        return self._images[idx]

    def __setitem__(self, idx, value):
        if not isinstance(value, SkyImage):
            raise TypeError("Item must be of 'SkyImage' type")
        value.sky += self._sky
        self._images[idx] = value
        self._update_bounding_polygon()

    def __delitem__(self, idx):
        del self._images[idx]
        if len(self._images) == 0:
            self._sky = 0.0
            self._id = None
        self._update_bounding_polygon()

    def __iter__(self):
        yield from self._images

    def insert(self, idx, value):
        """Inserts a `SkyImage` into the group."""
        if not isinstance(value, SkyImage):
            raise TypeError("Item must be of 'SkyImage' type")
        value.sky += self._sky
        self._images.insert(idx, value)
        self._update_bounding_polygon()

    def append(self, value):
        """Appends a `SkyImage` to the group."""
        if not isinstance(value, SkyImage):
            raise TypeError("Item must be of 'SkyImage' type")
        value.sky += self._sky
        self._images.append(value)
        self._update_bounding_polygon()

    def calc_sky(self, overlap=None, delta=True):
        """
        Compute sky background value.

        Parameters
        ----------
        overlap : SkyImage, SkyGroup, SphericalPolygon, list of tuples, \
None, optional
            Another `SkyImage`, `SkyGroup`,
            :py:class:`spherical_geometry.polygons.SphericalPolygon`, or
            a list of tuples of (RA, DEC) of vertices of a spherical
            polygon. This parameter is used to indicate that sky statistics
            should computed only in the region of intersection of *this*
            image with the polygon indicated by `overlap`. When `overlap` is
            `None`, sky statistics will be computed over the entire image.

        delta : bool, optional
            Should this function return absolute sky value or the difference
            between the computed value and the value of the sky stored in the
            `sky` property.

        Returns
        -------
        skyval : float, None
            Computed sky value (absolute or relative to the `sky` attribute).
            If there are no valid data to perform this computations (e.g.,
            because this image does not overlap with the image indicated by
            `overlap`), `skyval` will be set to `None`.

        npix : int
            Number of pixels used to compute sky statistics.

        polyarea : float
            Area (in srad) of the polygon that bounds data used to compute
            sky statistics.

        """

        if len(self._images) == 0:
            return None, 0, 0.0

        wght = 0
        area = 0.0

        if overlap is None:
            # compute minimum sky across all images in the group:
            wsky = None

            for image in self._images:
                # make sure all images have the same background:
                image.background = self._sky

                sky, npix, imarea = image.calc_sky(overlap=None, delta=delta)

                if sky is None:
                    continue

                if wsky is None or wsky > sky:
                    wsky = sky
                    wght = npix
                    area = imarea

            return wsky, wght, area

        # compute weighted sky in various overlaps:
        wsky = 0.0

        for image in self._images:
            # make sure all images have the same background:
            image.background = self._sky

            sky, npix, area1 = image.calc_sky(overlap=overlap, delta=delta)

            area += area1

            if sky is not None and npix > 0:
                pix_area = npix * image.pix_area
                wsky += sky * pix_area
                wght += pix_area

        if wght == 0.0 or area == 0.0:
            return None, wght, area
        else:
            return wsky / wght, wght, area



from enum import IntEnum

import numpy as np
import pytest
from astropy import units as u
from roman_datamodels.datamodels import RampModel, RefpixRefModel
from roman_datamodels.maker_utils import mk_ramp, mk_refpix

from romancal.refpix.data import Coefficients, Const, StandardView

RNG = np.random.default_rng(42)


class Dims(IntEnum):
    """
    Assumed data dimensions for dat (aid clarity of testing).
        - N_FRAMES: the number of frames used in the test data (arbitrary) choice
        - N_ROWS: the number of rows of the test data (the correction is
                  independent of this)
        - N_COLS: the number of columns of the test data (the correction is very
                  dependent on this)
    """

    N_FRAMES = 8
    N_ROWS = 12
    N_COLS = Const.N_COLUMNS + Const.CHAN_WIDTH
    N_CHAN = N_COLS // Const.CHAN_WIDTH
    PADDED_WIDTH = Const.CHAN_WIDTH + Const.PAD

    N_FFT_COLS = (N_ROWS * PADDED_WIDTH) // 2 + 1


@pytest.fixture(scope="module")
def data():
    return RNG.uniform(1, 100, size=(Dims.N_FRAMES, Dims.N_ROWS, Dims.N_COLS)).astype(
        np.float32
    )


@pytest.fixture(scope="module")
def datamodel(data):
    datamodel = mk_ramp(shape=(2, 2, 2))

    detector = data[:, :, : Const.N_COLUMNS]
    amp33 = data[:, :, Const.N_COLUMNS :]

    # Sanity check, that this is in line with the datamodel maker_utils
    assert detector.shape == (Dims.N_FRAMES, Dims.N_ROWS, Const.N_COLUMNS)
    assert amp33.shape == (Dims.N_FRAMES, Dims.N_ROWS, Const.CHAN_WIDTH)

    datamodel.data = u.Quantity(
        detector, unit=datamodel.data.unit, dtype=datamodel.data.dtype
    )
    datamodel.amp33 = u.Quantity(
        amp33, unit=datamodel.amp33.unit, dtype=datamodel.amp33.dtype
    )

    datamodel.border_ref_pix_left = datamodel.data[:, :, : Const.REF]
    datamodel.border_ref_pix_right = datamodel.data[:, :, -Const.REF :]

    return RampModel(datamodel)


@pytest.fixture(scope="module")
def standard(data):
    return StandardView(data)


@pytest.fixture(scope="module")
def channels(standard):
    return standard.channels


@pytest.fixture(scope="module")
def coeffs() -> Coefficients:
    coeffs = RNG.uniform(1, 100, size=(6, Const.N_DETECT_CHAN, Dims.N_FFT_COLS))

    coeffs = coeffs[:3, :, :] + 1.0j * coeffs[3:, :, :]

    return Coefficients(coeffs[0], coeffs[1], coeffs[2])


@pytest.fixture(scope="module")
def offset() -> np.ndarray:
    return RNG.uniform(0, 100, size=(Dims.N_ROWS, Dims.N_COLS)).astype(np.float32)


@pytest.fixture(scope="module")
def ref_pix_ref(coeffs):
    refpix_ref = mk_refpix(gamma=coeffs.gamma, zeta=coeffs.zeta, alpha=coeffs.alpha)

    return RefpixRefModel(refpix_ref)



from docutils.nodes import SkipNode, Text, caption, figure, raw, reference
from sphinx.roles import XRefRole

# Element classes


class page_ref(reference):
    pass


class num_ref(reference):
    pass


# Visit/depart functions


def skip_page_ref(self, node):
    raise SkipNode


def latex_visit_page_ref(self, node):
    self.body.append(f"\\pageref{{{node['refdoc']}:{node['reftarget']}}}")
    raise SkipNode


def latex_visit_num_ref(self, node):
    fields = node["reftarget"].split("#")
    if len(fields) > 1:
        label, target = fields
        ref_link = f"{node['refdoc']}:{target}"
        latex = f"\\hyperref[{ref_link}]{{{label} \\ref*{{{ref_link}}}}}"
        self.body.append(latex)
    else:
        self.body.append(f"\\ref{{{node['refdoc']}:{fields[0]}}}")

    raise SkipNode


def doctree_read(app, doctree):
    # first generate figure numbers for each figure
    env = app.builder.env
    figid_docname_map = getattr(env, "figid_docname_map", {})

    for figure_info in doctree.traverse(figure):
        for id in figure_info["ids"]:
            figid_docname_map[id] = env.docname

    env.figid_docname_map = figid_docname_map


def doctree_resolved(app, doctree, docname):
    i = 1
    figids = {}
    for figure_info in doctree.traverse(figure):
        if app.builder.name != "latex" and app.config.number_figures:
            for cap in figure_info.traverse(caption):
                cap[0] = Text(
                    "%s %d: %s" % (app.config.figure_caption_prefix, i, cap[0])
                )

        for id in figure_info["ids"]:
            figids[id] = i

        i += 1

    # replace numfig nodes with links
    if app.builder.name != "latex":
        for ref_info in doctree.traverse(num_ref):
            if "#" in ref_info["reftarget"]:
                label, target = ref_info["reftarget"].split("#")
                labelfmt = label + " %d"
            else:
                labelfmt = "%d"
                target = ref_info["reftarget"]

            if target not in figids:
                continue

            if app.builder.name == "html":
                target_doc = app.builder.env.figid_docname_map[target]
                link = f"{app.builder.get_relative_uri(docname, target_doc)}#{target}"
                html = (
                    f'<a class="pageref" href="{link}">{labelfmt %(figids[target])}</a>'
                )
                ref_info.replace_self(raw(html, html, format="html"))
            else:
                ref_info.replace_self(Text(labelfmt % (figids[target])))


def clean_env(app):
    app.builder.env.i = 1
    app.builder.env.figid_docname_map = {}


def setup(app):
    app.add_config_value("number_figures", True, True)
    app.add_config_value("figure_caption_prefix", "Figure", True)

    app.add_node(
        page_ref,
        text=(skip_page_ref, None),
        html=(skip_page_ref, None),
        latex=(latex_visit_page_ref, None),
    )

    app.add_role("page", XRefRole(nodeclass=page_ref))

    app.add_node(num_ref, latex=(latex_visit_num_ref, None))

    app.add_role("num", XRefRole(nodeclass=num_ref))

    app.connect("builder-inited", clean_env)
    app.connect("doctree-read", doctree_read)
    app.connect("doctree-resolved", doctree_resolved)



"""Functionality required by DMS outside of core functionality of the package

Example: Certain tests are required by DMS to log in a specific way. The
decorator for this is defined in this module.
"""

from stpipe import log as stpipe_log


def log_result(requirement, message, result):
    """Log individual test results that relate to a requirement

    Parameters
    ----------
    requirement : str
        The requirement being logged. I.e "DMS363"

    message : str
        Message describing what is being tested

    result : bool
        The result of the test
    """
    logger = stpipe_log.delegator.log
    result_text = "PASS" if result else "FAIL"
    log_msg = f"{requirement} MSG: {message}.......{result_text}"
    logger.info(log_msg)



"""Callback registry"""

from romancal.lib.signal_slot import Signal

__all__ = ["CallbackRegistry"]


class CallbackRegistry:
    """Callback registry"""

    def __init__(self):
        self.registry = dict()

    def add(self, event, callback):
        """Add a callback to an event"""
        try:
            signal = self.registry[event]
        except KeyError:
            signal = Signal()
        signal.connect(callback)
        self.registry[event] = signal

    def reduce(self, event, *args):
        """Perform a reduction on the event args

        Parameters
        ----------
        args : [arg[,...]]
            The args to filter

        Returns
        -------
        The reduced results.
        If no results can be determined,
        such as if no callbacks were registered,
        `None` is returned.

        Notes
        -----
        Each function is given the results of the previous
        function. As such, if the data has more than one
        object, the return of each function should be a tuple that can
        then be passed as arguments to the next function.

        There is no guarantee on order which the registered
        callbacks are made. Currently, the callbacks are in a list.
        Hence, the callbacks will be called in the order registered.

        """
        result = self.registry[event].reduce(*args)
        return result

    def add_decorator(self, event):
        """Add callbacks by decoration

        Parameters
        ----------
        event : str
            The name of event to attach the object to.
        """

        def decorator(func):
            self.add(event, func)
            return func

        return decorator

    __call__ = add_decorator



"""Public common step definition for OutlierDetection processing."""

from functools import partial
from pathlib import Path

from romancal.datamodels import ModelContainer
from romancal.outlier_detection import outlier_detection

from ..stpipe import RomanStep

__all__ = ["OutlierDetectionStep"]


class OutlierDetectionStep(RomanStep):
    """Flag outlier bad pixels and cosmic rays in DQ array of each input image.

    Input images can be listed in an input association file or already wrapped
    with a ModelContainer. DQ arrays are modified in place.

    Parameters
    -----------
    input_data : `~romancal.datamodels.container.ModelContainer`
        A `~romancal.datamodels.container.ModelContainer` object.

    """

    class_alias = "outlier_detection"

    # The members of spec needs to be a super-set of all parameters needed
    # by the various versions of the outlier_detection algorithms, and each
    # version will pick and choose what they need while ignoring the rest.
    spec = """
        weight_type = option('ivm','exptime',default='ivm') # Weighting type to use to create the median image
        pixfrac = float(default=1.0) # Fraction by which input pixels are shrunk before being drizzled onto the output image grid
        kernel = string(default='square') # Shape of the kernel used for flux distribution onto output images
        fillval = string(default='INDEF') # Value assigned to output pixels that have zero weight or no flux during drizzling
        nlow = integer(default=0) # The number of low values in each pixel stack to ignore when computing the median value
        nhigh = integer(default=0) # The number of high values in each pixel stack to ignore when computing the median value
        maskpt = float(default=0.7) # Percentage of weight image values below which they are flagged as bad pixels
        grow = integer(default=1) # The distance beyond the rejection limit for additional pixels to be rejected in an image
        snr = string(default='5.0 4.0') # The signal-to-noise values to use for bad pixel identification
        scale = string(default='1.2 0.7') # The scaling factor applied to derivative used to identify bad pixels
        backg = float(default=0.0) # User-specified background value to subtract during final identification step
        kernel_size = string(default='7 7') # Size of kernel to be used during resampling of the data
        save_intermediate_results = boolean(default=False) # Specifies whether or not to write out intermediate products to disk
        resample_data = boolean(default=True) # Specifies whether or not to resample the input images when performing outlier detection
        good_bits = string(default="~DO_NOT_USE+NON_SCIENCE")  # DQ bit value to be considered 'good'
        allowed_memory = float(default=None)  # Fraction of memory to use for the combined image
        in_memory = boolean(default=False) # Specifies whether or not to keep all intermediate products and datamodels in memory
    """  # noqa: E501

    def process(self, input_models):
        """Perform outlier detection processing on input data."""

        self.skip = False

        try:
            self.input_models = ModelContainer(input_models)
        except TypeError:
            self.log.warning(
                "Skipping outlier_detection - input cannot be parsed into a ModelContainer."
            )
            self.skip = True
            return input_models

        # validation
        if len(self.input_models) >= 2 and all(
            model.meta.exposure.type == "WFI_IMAGE" for model in self.input_models
        ):
            # Setup output path naming if associations are involved.
            asn_id = self.input_models.asn_table.get("asn_id", None)
            if asn_id is not None:
                _make_output_path = self.search_attr(
                    "_make_output_path", parent_first=True
                )
                self._make_output_path = partial(_make_output_path, asn_id=asn_id)

            detection_step = outlier_detection.OutlierDetection
            pars = {
                "weight_type": self.weight_type,
                "pixfrac": self.pixfrac,
                "kernel": self.kernel,
                "fillval": self.fillval,
                "nlow": self.nlow,
                "nhigh": self.nhigh,
                "maskpt": self.maskpt,
                "grow": self.grow,
                "snr": self.snr,
                "scale": self.scale,
                "backg": self.backg,
                "kernel_size": self.kernel_size,
                "save_intermediate_results": self.save_intermediate_results,
                "resample_data": self.resample_data,
                "good_bits": self.good_bits,
                "allowed_memory": self.allowed_memory,
                "in_memory": self.in_memory,
                "make_output_path": self.make_output_path,
                "resample_suffix": "i2d",
            }

            self.log.debug(
                f"Using {detection_step.__name__} class for outlier_detection"
            )

            # Set up outlier detection, then do detection
            step = detection_step(self.input_models, **pars)
            step.do_detection()

            state = "COMPLETE"

            if not self.save_intermediate_results:
                self.log.debug(
                    "The following files will be deleted since \
                    save_intermediate_results=False:"
                )
            for model in self.input_models:
                model.meta.cal_step["outlier_detection"] = state
                if not self.save_intermediate_results:
                    #  remove intermediate files found in
                    #  make_output_path() and the local dir
                    intermediate_files_paths = [
                        Path(self.make_output_path()).parent,
                        Path().cwd(),
                    ]
                    intermediate_files_suffixes = (
                        "*blot.asdf",
                        "*median.asdf",
                        f'*{pars.get("resample_suffix")}*.asdf',
                    )
                    for current_path in intermediate_files_paths:
                        for suffix in intermediate_files_suffixes:
                            for filename in current_path.glob(suffix):
                                filename.unlink()
                                self.log.debug(f"    {filename}")

        else:
            # if input can be parsed into a ModelContainer
            # but is not valid then log a warning message and
            # skip outlier detection step
            self.log.warning(
                "Skipping outlier_detection - at least two imaging observations are needed."
            )
            # set meta.cal_step.outlier_detection to SKIPPED
            for model in self.input_models:
                model.meta.cal_step["outlier_detection"] = "SKIPPED"
            self.skip = True

        return self.input_models
